import sys
import time

from .config import read_config
from .logger import ConsoleLogger

start_time = time.time()


def process_ini(config_filename):
    logger = ConsoleLogger()

    graph = read_config(
        config_filename,
        logger=logger,
    )
    graph.run(logger)


def run(argv):
    for config_filename in argv:
        process_ini(config_filename)


def main():
    run(sys.argv[1:])


if __name__ == "__main__":
    main()  # pragma: no cover



import gzip
from itertools import islice

import pandas as pd
from fqfa.fastq.fastq import parse_fastq_reads  # type: ignore

from countess import VERSION
from countess.core.parameters import BooleanParam, FloatParam
from countess.core.plugins import PandasInputFilesPlugin


def _file_reader(file_handle, min_avg_quality, row_limit=None, filename=""):
    for fastq_read in islice(parse_fastq_reads(file_handle), 0, row_limit):
        if fastq_read.average_quality() >= min_avg_quality:
            yield {"sequence": fastq_read.sequence, "header": fastq_read.header[1:], "filename": filename}


class LoadFastqPlugin(PandasInputFilesPlugin):
    """Load counts from one or more FASTQ files, by first building a dask dataframe of raw sequences
    with count=1 and then grouping by sequence and summing counts.  It supports counting
    in multiple columns."""

    name = "FASTQ Load"
    description = "Loads counts from FASTQ files containing either variant or barcodes"
    link = "https://countess-project.github.io/CountESS/included-plugins/#fastq-load"
    version = VERSION

    file_types = [("FASTQ", ".fastq"), ("FASTQ (gzipped)", ".gz")]

    parameters = {
        "min_avg_quality": FloatParam("Minimum Average Quality", 10),
        "header_column": BooleanParam("Header Column?", False),
        "filename_column": BooleanParam("Filename Column?", False),
        "group": BooleanParam("Group by Sequence?", True),
    }

    def read_file_to_dataframe(self, file_params, logger, row_limit=None):
        filename = file_params["filename"].value
        min_avg_quality = self.parameters["min_avg_quality"].value

        if filename.endswith(".gz"):
            with gzip.open(filename, mode="rt", encoding="utf-8") as fh:
                dataframe = pd.DataFrame(_file_reader(fh, min_avg_quality, row_limit, filename))
        else:
            with open(filename, "r", encoding="utf-8") as fh:
                dataframe = pd.DataFrame(_file_reader(fh, min_avg_quality, row_limit, filename))

        group_columns = ["sequence"]

        if self.parameters["header_column"].value:
            # find maximum common length of the 'header' field in this file
            for common_length in range(0, dataframe["header"].str.len().min() - 1):
                if dataframe["header"].str.slice(0, common_length + 1).nunique() > 1:
                    break
            if common_length > 0:
                dataframe["header"] = dataframe["header"].str.slice(0, common_length)
                group_columns.append("header")
        else:
            dataframe.drop(columns="header", inplace=True)

        if self.parameters["filename_column"].value:
            group_columns.append("filename")
        else:
            dataframe.drop(columns="filename", inplace=True)

        if self.parameters["group"].value:
            return dataframe.assign(count=1).groupby(group_columns).count()
        else:
            return dataframe



"""countess/gui/tree.py: handle presentation of the datagraph
tree diagram in a canvas."""

# XXX there is much here which should be rewritten. See:
# https://github.com/CountESS-Project/CountESS/issues/19
# for more discussion etc.

import math
import random
import re
import tkinter as tk
from enum import Enum, IntFlag
from functools import partial

from countess.core.pipeline import PipelineNode
from countess.gui.widgets import get_icon


def _limit(value, min_value, max_value):
    return max(min_value, min(max_value, value))


def _geometry(widget):
    return (
        widget.winfo_x(),
        widget.winfo_y(),
        widget.winfo_width(),
        widget.winfo_height(),
    )


# Weird that these aren't defined anywhere obvious?
# XXX note cross-platform mapping for modifiers and buttons
# https://wiki.tcl-lang.org/page/Modifier+Keys


class TkEventState(IntFlag):
    SHIFT = 1
    CAPS_LOCK = 2
    CONTROL = 4
    MOD1 = 8
    MOD2 = 16
    MOD3 = 32
    MOD4 = 64
    MOD5 = 128
    BUTTON1 = 256
    BUTTON2 = 512
    BUTTON3 = 1024
    BUTTON4 = 2048
    BUTTON5 = 4096


class TkCursors(Enum):
    HAND = "hand1"
    ARROWS = "fleur"
    PLUS = "plus"


DraggableMixinState = Enum("DraggableMixinState", ["READY", "DRAG_WAIT", "LINK_WAIT", "DRAGGING", "LINKING"])


class DraggableMixin:  # pylint: disable=R0903
    __state = DraggableMixinState.READY

    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self.bind("<Button-1>", self.__on_button, add=True)
        self.bind("<Button-2>", self.__on_button, add=True)
        self.bind("<Button-3>", self.__on_button, add=True)
        self.bind("<B1-Motion>", self.__on_motion, add=True)
        self.bind("<B2-Motion>", self.__on_motion, add=True)
        self.bind("<B3-Motion>", self.__on_motion, add=True)
        self.bind("<ButtonRelease-1>", self.__on_release, add=True)
        self.bind("<ButtonRelease-2>", self.__on_release, add=True)
        self.bind("<ButtonRelease-3>", self.__on_release, add=True)
        self["cursor"] = TkCursors.HAND.value

    def __on_button(self, event):
        if event.state & TkEventState.CONTROL or event.num > 1:
            self.__state = self.__state.LINK_WAIT
        else:
            self.__state = self.__state.DRAG_WAIT
        self.after(100, self.__on_timeout)

    def __place(self, event=None):
        x, y, w, h = _geometry(self)
        x = x + (event.x if event else (w // 2))
        y = y + (event.y if event else (h // 2))
        return {
            "relx": _limit(x / self.master.winfo_width(), 0.05, 0.95),
            "rely": _limit(y / self.master.winfo_height(), 0.05, 0.95),
            "anchor": "c",
        }

    def __on_timeout(self):
        if self.__state == self.__state.DRAG_WAIT:
            self.__state = self.__state.DRAGGING
            self["cursor"] = TkCursors.ARROWS.value
        elif self.__state == self.__state.LINK_WAIT:
            self.__state = self.__state.LINKING
            self.__ghost = tk.Frame(self.master)
            self.__ghost.place(self.__place())
            self.__ghost_line = ConnectingLine(self.master, self, self.__ghost, "red", True)
            self["cursor"] = TkCursors.PLUS.value

    def __on_motion(self, event):
        if self.__state == self.__state.DRAGGING:
            self.place(self.__place(event))
        elif self.__state == self.__state.LINKING:
            self.__ghost.place(self.__place(event))

    def __on_release(self, event):
        if self.__state == self.__state.LINKING:
            self.__ghost_line.destroy()
            self.__ghost.destroy()
            self.event_generate("<<GhostRelease>>", x=event.x, y=event.y)
        self.__state = self.__state.READY
        self["cursor"] = TkCursors.HAND.value


class FixedUnbindMixin:  # pylint: disable=R0903
    def unbind(self, seq, funcid=None):
        # widget.unbind(seq, funcid) doesn't actually work as documented. This is
        # my own take on the horrible hacks found at https://bugs.python.org/issue31485
        # and https://mail.python.org/pipermail/tkinter-discuss/2012-May/003152.html
        # Also quite horrible.  "I'm not proud, but I'm not tired either"
        if not funcid:
            for f in re.findall(r'^if {"\[(\w+).*$', self.bind(seq), flags=re.M):
                self.deletecommand(f)
            super().unbind(seq, None)
        else:
            self.bind(
                seq,
                re.sub(r'^if {"\[' + funcid + ".*$", "", self.bind(seq), flags=re.M),
            )
            self.deletecommand(funcid)


class ConnectingLine:
    line = None

    def __init__(self, canvas, widget1, widget2, color="black", switch=False):
        self.canvas = canvas
        self.widget1 = widget1
        self.widget2 = widget2
        self.color = color
        self.switch = switch

        # XXX I'm not totally in love with all this event
        # binding and unbinding and might replace it with
        # a GraphWrapper-level binding.

        self.widget1_bind = self.widget1.bind("<Configure>", self.update_line, add=True)
        self.widget2_bind = self.widget2.bind("<Configure>", self.update_line, add=True)
        self.canvas_bind = self.canvas.bind("<Configure>", self.update_line, add=True)
        self.update_line()

    def update_line(self, event=None):
        _xc, _yc, wc, hc = _geometry(self.canvas)

        # size factor to allow for different screen geometries.
        k = (wc + hc) / 100

        x1, y1, w1, h1 = _geometry(self.widget1)
        x2, y2, w2, h2 = _geometry(self.widget2)

        # special case for dragging invisible frames
        # XXX bit of a hack just to get it to look nice
        # based on the cursor size not `k`.
        if w2 == 1 and h2 == 1:
            x2, y2, w2, h2 = x2 - 16, y2 - 16, 32, 32

        # Control points set up a nice spline, and a little extra padding
        # on the destination end to allow for the arrow head.
        if wc > hc:
            if self.switch and x1 > x2:
                x1, y1, w1, h1, x2, y2, w2, h2 = x2, y2, w2, h2, x1, y1, w1, h1
            coords = (
                x1 + w1,
                y1 + h1 // 2,
                x1 + w1 + k,
                y1 + h1 // 2,
                x2 - k * 2,
                y2 + h2 // 2,
                x2,
                y2 + h2 // 2,
            )
        else:
            if self.switch and y1 > y2:
                x1, y1, w1, h1, x2, y2, w2, h2 = x2, y2, w2, h2, x1, y1, w1, h1
            coords = (
                x1 + w1 // 2,
                y1 + h1,
                x1 + w1 // 2,
                y1 + h1 + k,
                x2 + w2 // 2,
                y2 - k * 2,
                x2 + w2 // 2,
                y2,
            )

        arrowshape = (15, 20, 6) if k > 20 else (k * 2 / 3, k, k / 3)
        if self.line:
            self.canvas.coords(self.line, *coords)
            self.canvas.itemconfig(self.line, smooth=True, arrowshape=arrowshape)
        else:
            self.line = self.canvas.create_line(
                *coords,
                smooth=True,
                width=3,
                arrow="last",
                arrowshape=arrowshape,
                fill=self.color,
            )

    def destroy(self):
        self.canvas.delete(self.line)
        self.canvas.unbind("<Configure>", self.canvas_bind)
        self.widget1.unbind("<Configure>", self.widget1_bind)
        self.widget2.unbind("<Configure>", self.widget2_bind)


class FlippyCanvas(FixedUnbindMixin, tk.Canvas):
    """A canvas which flips all its children's X and Y
    coordinates when it goes from portrait to landscape.
    Place children with .place(relx=, rely=) for best results."""

    __flipped = False

    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self.bind("<Configure>", self.__on_configure)

    def __flip(self, width, height):
        for c in self.winfo_children():
            c.place(
                {
                    "x": c.place_info()["y"],
                    "y": c.place_info()["x"],
                    "relx": c.place_info()["rely"],
                    "rely": c.place_info()["relx"],
                }
            )
        # XXX should flip canvas items as well,
        # but I didn't actually need that so I didn't bother.

    def __on_configure(self, event):
        flipped = event.height > event.width
        if flipped != self.__flipped:
            self.__flip(event.width, event.height)
            self.__flipped = flipped


class DraggableLabel(DraggableMixin, FixedUnbindMixin, tk.Label):
    pass


class NodeWrapper(DraggableLabel):
    def update_node(self, node, vertical=False):
        input_bar = node.plugin and node.plugin.num_inputs == 0
        output_bar = node.plugin and node.plugin.num_outputs == 0
        if not input_bar and not output_bar:
            image = None
            compound = tk.NONE
        elif vertical:
            image = get_icon(self, "hbar")
            compound = tk.TOP if input_bar else tk.BOTTOM
        else:
            image = get_icon(self, "vbar")
            compound = tk.LEFT if input_bar else tk.RIGHT
        self.configure(text=node.name, image=image, compound=compound)


class GraphWrapper:
    selected_node = None
    highlight_rectangle = None

    def __init__(self, canvas, graph, node_select_callback):
        self.canvas = canvas
        self.graph = graph
        self.node_select_callback = node_select_callback

        self.labels = dict((node, self.label_for_node(node)) for node in graph.nodes)
        self.lines = dict((node, self.lines_for_node(node)) for node in graph.nodes)
        self.lines_lookup = dict(
            (connecting_line.line, (connecting_line, child_node, parent_node))
            for child_node, lines_for_child in self.lines.items()
            for parent_node, connecting_line in lines_for_child.items()
        )

        self.canvas.bind("<Button-1>", self.on_canvas_button)
        self.canvas.bind("<Button-2>", self.on_canvas_button)
        self.canvas.bind("<Button-3>", self.on_canvas_button)
        self.canvas.bind("<Motion>", self.on_canvas_motion)
        self.canvas.bind("<Leave>", self.on_canvas_leave)
        self.canvas.bind("<Key-Delete>", self.on_canvas_delete)
        self.canvas.bind("<Key-BackSpace>", self.on_canvas_delete)

    def label_for_node(self, node):
        label = NodeWrapper(self.canvas, wraplength=125, cursor="hand1", takefocus=True)
        if not node.position:
            node.position = (random.random() * 0.8 + 0.1, random.random() * 0.8 + 0.1)
        # XXX should be more elegant way of answering the question "are we flipped?"
        if self.canvas.winfo_width() >= self.canvas.winfo_height():
            label.place({"relx": node.position[0], "rely": node.position[1], "anchor": "c"})
        else:
            label.place({"relx": node.position[1], "rely": node.position[0], "anchor": "c"})

        label.bind("<Button-1>", partial(self.on_mousedown, node), add=True)
        label.bind("<Configure>", partial(self.on_configure, node, label), add=True)
        label.bind("<<GhostRelease>>", partial(self.on_ghost_release, node), add=True)
        label.bind("<Key-Delete>", partial(self.on_delete, node), add=True)
        label.bind("<Key-BackSpace>", partial(self.on_delete, node), add=True)
        label.bind("<Enter>", partial(self.on_enter, node), add=True)
        label.bind("<Leave>", partial(self.on_leave, node), add=True)

        return label

    def lines_for_node(self, node):
        return dict(
            (
                parent_node,
                ConnectingLine(self.canvas, self.labels[parent_node], self.labels[node]),
            )
            for parent_node in node.parent_nodes
        )

    def highlight_node(self, node):
        """Highlight the current node in orange"""
        self.selected_node = node
        for label in self.labels.values():
            label["bg"] = "#DDD"
        if node:
            self.labels[node]["bg"] = "orange"

    def on_mousedown(self, node, event):
        """Button-1 click selects a node to view"""
        if self.highlight_rectangle is not None:
            self.canvas.delete(self.highlight_rectangle)
            self.highlight_rectangle = None
        self.highlight_node(node)
        self.node_select_callback(node)

    def on_configure(self, node, label, event):
        """Stores the updated position of the label in node.position"""
        height = self.canvas.winfo_height()
        width = self.canvas.winfo_width()

        xx = float(label.place_info()["relx"]) * width
        yy = float(label.place_info()["rely"]) * height
        node.position = self.new_node_position(xx, yy)
        label.update_node(node, width < height)

        # Adapt label sizes to suit the window size, as best we can ...
        # XXX very arbitrary and definitely open to tweaking
        if height > width:
            label_max_width = max(width // 9, 25)
            label_font_size = int(math.sqrt(width) / 3)
        else:
            label_max_width = max(width // 20, 16)
            label_font_size = int(math.sqrt(width) / 5)
        label["wraplength"] = label_max_width
        label["font"] = ("TkDefaultFont", label_font_size)

    def on_enter(self, node, event):
        """Mouse has entered a label"""
        # don't outline labels while dragging
        if event.state & TkEventState.BUTTON1:
            return
        self.labels[node].focus()
        # hide any highlighted lines
        self.on_canvas_leave(event)
        x, y, w, h = _geometry(self.labels[node])
        if self.highlight_rectangle:
            self.canvas.delete(self.highlight_rectangle)
        self.highlight_rectangle = self.canvas.create_rectangle(x - 3, y - 3, x + w + 3, y + h + 3, fill="red", width=0)

    def on_leave(self, node, event):
        if self.highlight_rectangle is not None:
            self.canvas.delete(self.highlight_rectangle)
            self.highlight_rectangle = None
        self.canvas.focus_set()

    def on_canvas_motion(self, event):
        """Show a preview of line selection when the cursor is over line(s)"""
        items = self.canvas.find_overlapping(event.x - 10, event.y - 10, event.x + 10, event.y + 10)
        for connecting_line, _, _ in self.lines_lookup.values():
            color = "red" if connecting_line.line in items else "black"
            self.canvas.itemconfig(connecting_line.line, fill=color)

    def on_canvas_leave(self, event):
        """Called when the mouse leaves the canvas *OR* the mouse enters
        any of the labels, to clear the line selection preview."""
        for connecting_line, _, _ in self.lines_lookup.values():
            self.canvas.itemconfig(connecting_line.line, fill="black")

    def new_node_position(self, x, y):
        flipped = self.canvas.winfo_height() > self.canvas.winfo_width()
        xp = _limit(x / self.canvas.winfo_width(), 0.05, 0.95)
        yp = _limit(y / self.canvas.winfo_height(), 0.05, 0.95)
        return (yp, xp) if flipped else (xp, yp)

    def on_canvas_button(self, event):
        """Control-Click or Secondary Click to create a new node, if it is created on top of a line
        that line is broken and the node is included."""

        if event.num == 1 and not event.state & TkEventState.CONTROL:
            return

        items = self.canvas.find_overlapping(event.x - 10, event.y - 10, event.x + 10, event.y + 10)

        # XXX creating a new node every time you click on the background
        # is proving quite annoying.  Lets see how it is to go without it.
        # (you can still create a new node by button-3-dragging from an
        # existing node, and if you really want a disconnected graph you can
        # delete the link to the new node!)
        # if not items:
        #    return

        position = self.new_node_position(event.x, event.y)
        new_node = self.add_new_node(position)

        for item in items:
            _, child_node, parent_node = self.lines_lookup[item]
            self.del_parent(parent_node, child_node)
            self.add_parent(new_node, child_node)
            self.add_parent(parent_node, new_node)

    def on_canvas_delete(self, event):
        """Delete key on canvas: delete line(s)."""
        items = self.canvas.find_overlapping(event.x - 10, event.y - 10, event.x + 10, event.y + 10)
        for item in items:
            _, child_node, parent_node = self.lines_lookup[item]
            self.del_parent(parent_node, child_node)

    def on_delete(self, node, event):
        """<Delete> disconnects a node from the graph, connects it parents to its children,
        and deletes the node.  <Shift-Delete> removes and deletes the node, but doesn't
        reconnect parents and children.  <Ctrl-Delete> disconnects the node but doesn't
        delete it."""

        self.on_leave(node, event)

        parent_nodes = list(node.parent_nodes)
        child_nodes = list(node.child_nodes)
        for line in self.lines[node].values():
            line.destroy()
        for parent_node in parent_nodes:
            node.del_parent(parent_node)
        for child_node in child_nodes:
            child_node.del_parent(node)
            self.lines[child_node].pop(node).destroy()
            if not event.state & TkEventState.SHIFT:
                for parent_node in parent_nodes:
                    self.add_parent(parent_node, child_node)
        if not event.state & TkEventState.CONTROL:
            self.graph.nodes.remove(node)
            del self.labels[node]
            del self.lines[node]
            event.widget.destroy()

        if len(self.graph.nodes) == 0:
            self.add_new_node(select=True)
        elif node == self.selected_node:
            # arbitrarily pick another node to show
            new_node = parent_nodes[0] if parent_nodes else child_nodes[0] if child_nodes else list(self.graph.nodes)[0]
            self.highlight_node(new_node)
            self.node_select_callback(new_node)

    def find_node_at_position(self, x, y):
        for node, label in self.labels.items():
            nx, ny, nw, nh = _geometry(label)
            if (nx <= x <= nx + nw) and (ny <= y <= ny + nh):
                return node
        return None

    def add_new_node(self, position=(0.5, 0.5), select=True):
        new_node = PipelineNode(name=f"NEW {len(self.graph.nodes)+1}", position=position)
        self.graph.add_node(new_node)
        self.labels[new_node] = self.label_for_node(new_node)
        self.labels[new_node].update()
        self.lines[new_node] = {}
        if select:
            self.highlight_node(new_node)
            self.node_select_callback(new_node)
        return new_node

    def on_ghost_release(self, start_node, event):
        xl, yl, _wl, _hl = _geometry(event.widget)
        flipped = self.canvas.winfo_height() > self.canvas.winfo_width()
        other_node = self.find_node_at_position(event.x + xl, event.y + yl)
        if other_node is None:
            position = self.new_node_position(event.x + xl, event.y + yl)
            other_node = self.add_new_node(position)
        elif other_node == start_node:
            return
        elif start_node in other_node.parent_nodes:
            other_node.del_parent(start_node)
            return
        elif other_node in start_node.parent_nodes:
            start_node.del_parent(other_node)
            return

        if start_node.is_ancestor_of(other_node):
            self.add_parent(start_node, other_node)
        elif other_node.is_ancestor_of(start_node):
            self.add_parent(other_node, start_node)
        elif (flipped and event.y > 0) or (not flipped and event.x > 0):
            self.add_parent(start_node, other_node)
        else:
            self.add_parent(other_node, start_node)

        self.highlight_node(other_node)
        self.node_select_callback(other_node)

    def add_parent(self, parent_node, child_node):
        if (
            (not parent_node.plugin or parent_node.plugin.num_outputs)
            and (not child_node.plugin or child_node.plugin.num_inputs)
            and parent_node not in child_node.parent_nodes
        ):
            child_node.add_parent(parent_node)
            connecting_line = ConnectingLine(self.canvas, self.labels[parent_node], self.labels[child_node])
            self.lines[child_node][parent_node] = connecting_line
            self.lines_lookup[connecting_line.line] = (
                connecting_line,
                child_node,
                parent_node,
            )

    def del_parent(self, parent_node, child_node):
        connecting_line = self.lines[child_node].pop(parent_node)
        del self.lines_lookup[connecting_line.line]
        connecting_line.destroy()
        child_node.del_parent(parent_node)

    def node_changed(self, node):
        """Called when something external updates the node's name, status
        or configuration."""
        flipped = self.canvas.winfo_width() >= self.canvas.winfo_height()
        self.labels[node].update_node(node, not flipped)

    def destroy(self):
        for node_lines in self.lines.values():
            for line in node_lines.values():
                line.destroy()
        for label in self.labels.values():
            label.destroy()



import re
from typing import Iterable, Optional

import pandas as pd

from countess import VERSION
from countess.core.logger import Logger
from countess.core.parameters import (
    ArrayParam,
    BooleanParam,
    ColumnChoiceParam,
    DataTypeChoiceParam,
    IntegerParam,
    MultiParam,
    StringParam,
)
from countess.core.plugins import PandasInputFilesPlugin, PandasTransformSingleToTuplePlugin


class RegexToolPlugin(PandasTransformSingleToTuplePlugin):
    name = "Regex Tool"
    description = "Apply regular expressions to a column to make new column(s)"
    link = "https://countess-project.github.io/CountESS/included-plugins/#regex-tool"
    version = VERSION

    parameters = {
        "column": ColumnChoiceParam("Input Column"),
        "regex": StringParam("Regular Expression", ".*"),
        "output": ArrayParam(
            "Output Columns",
            MultiParam(
                "Col",
                {
                    "name": StringParam("Column Name"),
                    "datatype": DataTypeChoiceParam(
                        "Column Type",
                        "string",
                    ),
                },
            ),
        ),
        "drop_column": BooleanParam("Drop Column", False),
        "drop_unmatch": BooleanParam("Drop Unmatched Rows", False),
        "multi": BooleanParam("Multi Match", False),
    }

    compiled_re = None

    def prepare(self, sources: list[str], row_limit: Optional[int] = None):
        super().prepare(sources, row_limit)
        self.compiled_re = re.compile(self.parameters["regex"].value)

    def process_dataframe(self, dataframe: pd.DataFrame, logger: Logger) -> Optional[pd.DataFrame]:
        assert isinstance(self.parameters["output"], ArrayParam)
        df = super().process_dataframe(dataframe, logger)
        if df is None:
            return None

        if self.parameters["drop_column"].value:
            column_name = self.parameters["column"].value
            if column_name in df.columns:
                df = df.drop(columns=column_name)
            else:
                # XXX maybe set up a helper function for this
                try:
                    df = df.reset_index(column_name, drop=True)
                except KeyError:
                    pass

        return df

    def process_value(self, value: str, logger: Logger) -> Optional[Iterable]:
        assert self.compiled_re is not None
        assert isinstance(self.parameters["output"], ArrayParam)
        if value is not None:
            try:
                if self.parameters["multi"].value:
                    return self.compiled_re.findall(str(value))
                else:
                    if match := self.compiled_re.match(str(value)):
                        return [
                            op.datatype.cast_value(val) for op, val in zip(self.parameters["output"], match.groups())
                        ]
                    else:
                        pass
                        # logger.info(f"{repr(value)} didn't match")
            except (TypeError, ValueError) as exc:
                logger.exception(exc)

        # If dropping unmatched values, return a simple None which will
        # be filtered out in series_to_dataframe below, otherwise return
        # a tuple of Nones which will fill in the unmatched row.

        if self.parameters["drop_unmatch"].value:
            return None
        else:
            return [None] * self.compiled_re.groups

    def series_to_dataframe(self, series: pd.Series) -> pd.DataFrame:
        # Unmatched rows return a single None, so we can easily drop
        # them out before doing further processing
        if self.parameters["drop_unmatch"].value:
            series.dropna(inplace=True)

        if self.parameters["multi"].value:
            series = series.explode()

        return super().series_to_dataframe(series)


class RegexReaderPlugin(PandasInputFilesPlugin):
    name = "Regex Reader"
    description = "Loads arbitrary data from line-delimited files"
    additional = """Applies a regular expression to each line to extract fields.
        If you're trying to read generic CSV or TSV files, use the CSV plugin
        instead as it handles escaping correctly."""
    link = "https://countess-project.github.io/CountESS/included-plugins/#regex-reader"
    version = VERSION

    file_types = [("CSV", "*.csv"), ("TXT", "*.txt")]

    parameters = {
        "regex": StringParam("Regular Expression", "(.*)"),
        "skip": IntegerParam("Skip Lines", 0),
        "output": ArrayParam(
            "Output Columns",
            MultiParam(
                "Col",
                {
                    "name": StringParam("Column Name"),
                    "datatype": DataTypeChoiceParam(
                        "Column Type",
                        "string",
                    ),
                },
            ),
        ),
    }

    def read_file_to_dataframe(self, file_params, logger, row_limit=None):
        pdfs = []

        compiled_re = re.compile(self.parameters["regex"].value)

        while compiled_re.groups > len(self.parameters["output"].params):
            self.parameters["output"].add_row()

        output_parameters = list(self.parameters["output"])[: compiled_re.groups]
        columns = [p.name.value or f"column_{n+1}" for n, p in enumerate(output_parameters)]

        records = []
        with open(file_params["filename"].value, "r", encoding="utf-8") as fh:
            for num, line in enumerate(fh):
                if num < self.parameters["skip"].value:
                    continue
                match = compiled_re.match(line)
                if match:
                    records.append((output_parameters[n].datatype.cast_value(g) for n, g in enumerate(match.groups())))
                else:
                    logger.warning(f"Row {num+1} did not match", detail=line)
                if row_limit is not None:
                    if len(records) >= row_limit or num > 100 * row_limit:
                        break
                elif len(records) >= 100000:
                    pdfs.append(pd.DataFrame.from_records(records, columns=columns))
                    records = []

        if len(records) > 0:
            pdfs.append(pd.DataFrame.from_records(records, columns=columns))

        if len(pdfs) == 0:
            return pd.DataFrame([], columns=columns)

        df = pd.concat(pdfs)
        return df

    def load_file(self, file_number: int, logger: Logger, row_limit: Optional[int] = None) -> Iterable:
        assert isinstance(self.parameters["files"], ArrayParam)
        file_params = self.parameters["files"][file_number]
        yield self.read_file_to_dataframe(file_params, logger, row_limit)



import re
from typing import Optional

import pandas as pd

from countess import VERSION
from countess.core.logger import Logger
from countess.core.parameters import BooleanParam, ColumnChoiceParam, ColumnOrNoneChoiceParam, IntegerParam, StringParam
from countess.core.plugins import PandasTransformDictToDictPlugin


class HgvsParserPlugin(PandasTransformDictToDictPlugin):
    name = "HGVS Parser"
    description = "Parse HGVS strings"

    version = VERSION

    parameters = {
        "column": ColumnChoiceParam("Input Column"),
        "guides_col": ColumnOrNoneChoiceParam("Guide(s) Column"),
        "guides_str": StringParam("Guide(s)"),
        "max_var": IntegerParam("Maximum Variations", 1),
        "split": BooleanParam("Split Output", False),
        "multi": BooleanParam("Multiple rows", False),
    }

    def process_dict(self, data: dict, logger: Logger):
        assert isinstance(self.parameters["guides_col"], ColumnOrNoneChoiceParam)
        try:
            value = data[self.parameters["column"].value]
        except KeyError:
            return None

        if value is None:
            return None

        output = {}

        guides = []
        if self.parameters["guides_col"].is_not_none():
            guides += str(data[self.parameters["guides_col"].value]).split(";")
        if self.parameters["guides_str"].value:
            guides += self.parameters["guides_str"].value.split(";")

        if m := re.match(r"([\w.]+):([ncg].)(.*)", value):
            output["reference"] = m.group(1)
            output["prefix"] = m.group(2)
            value = m.group(3)

        if value == "=":
            variations = []
        else:
            if value.startswith("[") and value.endswith("]"):
                value = value[1:-1]
            variations = value.split(";")

        for n, g in enumerate(guides, 1):
            output[f"guide_{n}"] = g in variations

        max_variations = self.parameters["max_var"].value
        variations = [v for v in variations if v not in guides]
        if len(variations) > max_variations:
            return None

        output_vars: list[Optional[str]] = [None] * max_variations
        output_locs: list[Optional[str]] = [None] * max_variations
        for n, v in enumerate(variations):
            if self.parameters["split"].value:
                if m := re.match(r"([\d_]+)(.*)", v):
                    output_locs[n] = m.group(1)
                    output_vars[n] = m.group(2)
                    continue
            output_vars[n] = v

        if self.parameters["multi"].value:
            output["var"] = output_vars
            if self.parameters["split"].value:
                output["loc"] = output_locs
        else:
            for n, (var, loc) in enumerate(zip(output_vars, output_locs), 1):
                output[f"var_{n}"] = var
                if self.parameters["split"].value:
                    output[f"loc_{n}"] = loc

        return output

    def series_to_dataframe(self, series: pd.Series) -> pd.DataFrame:
        dataframe = super().series_to_dataframe(series)

        if self.parameters["multi"].value:
            if self.parameters["split"].value:
                dataframe = dataframe.explode(["var", "loc"])
            else:
                dataframe = dataframe.explode(["var"])

        return dataframe



import importlib.metadata
from unittest.mock import patch

import countess
from countess.core.plugins import BasePlugin, get_plugin_classes

empty_entry_points_dict = {"countess_plugins": []}

invalid_entry_points_dict = {
    "countess_plugins": [importlib.metadata.EntryPoint(name="test", value="mockplugin", group="countess_plugins")]
}


class NoParentPlugin:
    pass


noparent_entry_points_dict = {
    "countess_plugins": [importlib.metadata.EntryPoint(name="test", value="NoParentPlugin", group="countess_plugins")]
}


def test_get_plugin_classes_invalid(caplog):
    with patch("importlib.metadata.entry_points", lambda: invalid_entry_points_dict):
        get_plugin_classes()
        assert "could not be loaded" in caplog.text


def test_get_plugin_classes_wrongparent(caplog):
    with patch("importlib.metadata.entry_points", lambda: noparent_entry_points_dict):
        with patch("importlib.metadata.EntryPoint.load", lambda x: NoParentPlugin):
            get_plugin_classes()
            assert "not a valid CountESS plugin" in caplog.text



import functools
from typing import Dict, List, Optional

import numpy as np
import pandas as pd

from countess import VERSION
from countess.core.logger import Logger
from countess.core.parameters import ChoiceParam, PerColumnArrayParam
from countess.core.plugins import PandasProcessPlugin
from countess.utils.pandas import get_all_columns


def _product(iterable):
    return functools.reduce(lambda x, y: x * y, iterable, 1)


class PivotPlugin(PandasProcessPlugin):
    """Groups a Pandas Dataframe by an arbitrary column and rolls up rows"""

    name = "Pivot Tool"
    description = """Groups a dataframe and pivots column values into columns.
        Expanded column values are duplicated for each combination of pivot values.
        Missing values default to 0, and duplicate values are summed."""
    version = VERSION
    link = "https://countess-project.github.io/CountESS/included-plugins/#pivot-tool"

    parameters = {
        "columns": PerColumnArrayParam(
            "Columns", ChoiceParam("Role", "Drop", choices=["Index", "Pivot", "Expand", "Drop"])
        ),
    }

    input_columns: Dict[str, np.dtype] = {}

    dataframes: List[pd.DataFrame] = []

    def prepare(self, sources: List[str], row_limit: Optional[int] = None):
        self.input_columns = {}

    def process(self, data: pd.DataFrame, source: str, logger: Logger):
        assert isinstance(self.parameters["columns"], PerColumnArrayParam)
        self.input_columns.update(get_all_columns(data))

        column_parameters = list(zip(self.input_columns, self.parameters["columns"]))
        index_cols = [col for col, param in column_parameters if param.value == "Index"]
        pivot_cols = [col for col, param in column_parameters if param.value == "Pivot"]
        expand_cols = [col for col, param in column_parameters if param.value == "Expand"]

        if not index_cols:
            logger.warning("No columns to index!")

        if not expand_cols:
            logger.warning("No columns to expand!")

        if not pivot_cols:
            logger.error("No columns to pivot on!")
            return []

        n_pivot = _product(data[pc].nunique() for pc in pivot_cols) * len(expand_cols)
        if n_pivot > 200:
            pivot_cols_str = ", ".join(pivot_cols)
            logger.error(f"Too many pivot combinations on {pivot_cols_str} ({n_pivot})")
            return []

        df = pd.pivot_table(
            data,
            values=expand_cols,
            index=index_cols,
            columns=pivot_cols,
            aggfunc="sum",
            fill_value=0,
        )

        if isinstance(df.columns, pd.MultiIndex):
            # Clean up MultiIndex names ... XXX until such time as CountESS supports them
            df.columns = [
                "__".join([f"{cn}_{cv}" if cn else cv for cn, cv in zip(df.columns.names, cc)]) for cc in df.columns
            ]  # type: ignore

        self.dataframes.append(df)
        return []

    def finalize(self, logger: Logger):
        assert isinstance(self.parameters["columns"], PerColumnArrayParam)
        column_parameters = list(zip(self.input_columns, self.parameters["columns"]))
        index_cols = [col for col, param in column_parameters if param.value == "Index"]
        if self.dataframes:
            df = pd.concat(self.dataframes)
            self.dataframes = []
            if index_cols:
                df = df.groupby(by=index_cols, group_keys=True).sum()
            yield df

        for p in self.parameters.values():
            p.set_column_choices(self.input_columns.keys())



import builtins
import math
from types import CodeType, NoneType

import numpy as np
import pandas as pd

from countess import VERSION
from countess.core.logger import Logger
from countess.core.parameters import BooleanParam, TextParam
from countess.core.plugins import PandasTransformDictToDictPlugin

# XXX pretty sure this is a job for ast.parse rather than just
# running compile() and exec() but that can wait.
# Builtins are restricted but there's still plenty of things which
# could go wrong here.

# These types will get copied to columns, anything else
# (eg: classes, methods, functions) won't.

SIMPLE_TYPES = set((bool, int, float, str, tuple, list, NoneType))

SAFE_BUILTINS = {
    x: builtins.__dict__[x]
    for x in "abs all any ascii bin bool bytearray bytes chr complex dict divmod "
    "enumerate filter float format frozenset hash hex id int len list map max min "
    "oct ord pow range reversed round set slice sorted str sum tuple type zip".split()
}
MATH_FUNCTIONS = {k: v for k, v in math.__dict__.items() if not k.startswith("__")}


class PythonPlugin(PandasTransformDictToDictPlugin):
    name = "Python Code"
    description = "Apply python code to each row."
    additional = """
        Columns are mapped to local variables and back.
        If you assign to a variable called "__filter",
        only rows where that value is true will be kept.
    """

    version = VERSION

    parameters = {
        "code": TextParam("Python Code"),
        "dropna": BooleanParam("Drop Null Columns?"),
    }

    code_object = None
    code_globals = {
        "__builtins__": SAFE_BUILTINS,
        **MATH_FUNCTIONS,
    }

    def process_dict(self, data: dict, logger: Logger):
        assert isinstance(self.parameters["code"], TextParam)
        assert isinstance(self.code_object, CodeType)

        try:
            exec(self.code_object, self.code_globals, data)  # pylint: disable=exec-used
        except Exception as exc:  # pylint: disable=broad-exception-caught
            logger.exception(exc)

        return dict((k, v) for k, v in data.items() if type(v) in SIMPLE_TYPES or isinstance(v, np.generic))

    def process_dataframe(self, dataframe: pd.DataFrame, logger: Logger) -> pd.DataFrame:
        """Override parent class because we a) want to reset
        the indexes so we can use their values easily and
        b) we don't need to merge afterwards"""

        # XXX cache this?
        self.code_object = compile(self.parameters["code"].value, "<PythonPlugin>", mode="exec")

        dataframe = dataframe.reset_index(drop=dataframe.index.names == [None])
        series = self.dataframe_to_series(dataframe, logger)
        dataframe = self.series_to_dataframe(series)

        if "__filter" in dataframe.columns:
            dataframe = dataframe.query("__filter").drop(columns="__filter")

        if self.parameters["dropna"].value:
            dataframe.dropna(axis=1, how="all", inplace=True)

        return dataframe



# TK based GUI for CountESS
import math
import tkinter as tk
from functools import partial
from tkinter import filedialog, ttk
from typing import Mapping, MutableMapping, Optional

import numpy as np

from ..core.parameters import (
    ArrayParam,
    BaseParam,
    BooleanParam,
    ChoiceParam,
    ColumnOrStringParam,
    FileArrayParam,
    FileParam,
    FileSaveParam,
    MultiParam,
    SimpleParam,
    TabularMultiParam,
    TextParam,
)
from ..core.plugins import BasePlugin
from .widgets import BooleanCheckbox, add_button, delete_button


def is_nan(v):
    return v is None or v is np.nan or (isinstance(v, float) and math.isnan(v))


class ParameterWrapper:
    # XXX it would probably be better to have ParameterWrapper classes per Parameter class
    # given the ridiculous amount of if/elif going on in here.

    # XXX actually refactoring this whole mess of nested wrappers into a single TreeviewEditor
    # might make more sense (now I've worked out how to make a TreeviewEditor)

    def __init__(  # pylint: disable=R0912,R0915
        self,
        tk_parent: tk.Widget,
        parameter: BaseParam,
        callback=None,
        delete_callback=None,
        level=0,
    ):
        self.tk_parent = tk_parent
        self.parameter = parameter
        self.callback = callback
        self.button = None
        self.level = level
        self.subwrapper_buttons: list[tk.Button] = []
        self.column_labels: list[tk.Label] = []

        self.subwrappers: Mapping[BaseParam, ParameterWrapper] = {}

        self.var: Optional[tk.Variable] = None
        self.entry: Optional[tk.Widget] = None
        self.label: Optional[tk.Widget] = None
        self.row_labels: list[tk.Widget] = []

        if isinstance(parameter, ArrayParam):
            self.label = None
        else:
            self.label = tk.Label(tk_parent, text=parameter.label)

        if isinstance(parameter, ChoiceParam):
            self.var = tk.StringVar(tk_parent, value=parameter.value)
            self.entry = ttk.Combobox(tk_parent, textvariable=self.var)
            self.entry["values"] = parameter.choices or [""]
            if isinstance(parameter, ColumnOrStringParam):
                self.entry.bind("<Key>", self.combobox_set)
                self.entry["state"] = "normal"
            else:
                self.entry["state"] = "readonly"
        elif isinstance(parameter, BooleanParam):
            self.entry = BooleanCheckbox(tk_parent, command=self.toggle_checkbox_callback)
            self.set_checkbox_value()
        elif isinstance(parameter, FileParam):
            self.entry = tk.Label(tk_parent, text=parameter.value)
            self.button = tk.Button(tk_parent, text="Select", width=3, command=self.change_file_callback)
        elif isinstance(parameter, TextParam):
            # tk.Text widget doesn't have a variable, for whatever reason,
            # so we use a different method!
            # XXX is this a simpler way to handle other kinds of fields too?
            self.entry = tk.Text(tk_parent, height=10)
            self.entry.insert("1.0", parameter.value)
            if parameter.read_only:
                self.entry["state"] = tk.DISABLED
            else:
                self.entry.bind("<<Modified>>", self.widget_modified_callback)
        elif isinstance(parameter, SimpleParam):
            self.var = tk.StringVar(tk_parent, value=parameter.value)
            self.entry = tk.Entry(tk_parent, textvariable=self.var)
            if parameter.read_only:
                self.entry["state"] = tk.DISABLED

        elif (
            isinstance(parameter, ArrayParam) and self.level == 0 and not isinstance(parameter.param, TabularMultiParam)
        ):
            self.entry = tk.Frame(tk_parent)
            self.entry.columnconfigure(0, weight=1)
            drc = self.delete_row_callback if not parameter.read_only else None
            self.update_subwrappers_framed(parameter.params, drc)
            if not parameter.read_only:
                self.button = tk.Button(
                    self.entry,
                    text=f"Add {parameter.param.label}",
                    command=self.add_row_callback,
                )
                self.button.grid(row=len(parameter.params) + 1)

        elif isinstance(parameter, ArrayParam) and (self.level < 3 or isinstance(parameter.param, TabularMultiParam)):
            label_frame_label = tk.Frame(tk_parent)
            tk.Label(label_frame_label, text=parameter.label).grid(row=0, column=0, padx=5)
            self.entry = tk.LabelFrame(tk_parent, labelwidget=label_frame_label, padx=10, pady=5)
            if not parameter.read_only:
                self.button = add_button(label_frame_label, command=self.add_row_callback)
                self.button.grid(row=0, column=1, padx=10)

            drc = self.delete_row_callback if not parameter.read_only else None
            if isinstance(parameter.param, MultiParam):
                self.update_subwrappers_tabular(parameter.params, drc)
            else:
                self.entry.columnconfigure(0, weight=0)
                self.entry.columnconfigure(1, weight=0)
                self.entry.columnconfigure(2, weight=1)

                # XXX hack because the empty labelframe collapses
                # for some reason.
                tk.Label(self.entry, text="").grid()

                self.update_subwrappers(parameter.params, drc)

        elif isinstance(parameter, (ArrayParam, MultiParam)):
            self.entry = tk.Frame(tk_parent)
            self.entry.columnconfigure(0, weight=0)
            self.entry.columnconfigure(1, weight=0)
            self.entry.columnconfigure(2, weight=1)
            if isinstance(parameter, ArrayParam):
                drc = self.delete_row_callback if not parameter.read_only else None
                self.update_subwrappers(parameter.params, drc)
                self.button = add_button(tk_parent, command=self.add_row_callback)
            else:
                self.update_subwrappers(parameter.params.values(), None)
        else:
            raise NotImplementedError(f"Unknown parameter type {parameter}")

        if self.var:
            self.var.trace("w", self.value_changed_callback)

        # XXX hang on, what if it's an array in an array?
        if not parameter.read_only and delete_callback and not self.button:
            self.button = delete_button(tk_parent, command=lambda: delete_callback(self))

        if not isinstance(parameter, BooleanParam):
            self.entry.grid(sticky=tk.EW, padx=10, pady=5)

    def update(self):
        if self.parameter.hide:
            self.entry["fg"] = self.entry["bg"]
        else:
            self.entry["fg"] = None

        if (
            isinstance(self.parameter, ArrayParam)
            and self.level == 0
            and not isinstance(self.parameter.param, TabularMultiParam)
        ):
            self.update_subwrappers_framed(self.parameter.params, self.delete_row_callback)
            if self.button:
                self.button.grid(row=len(self.parameter.params) + 1, padx=10)
        elif isinstance(self.parameter, ArrayParam) and isinstance(self.parameter.param, MultiParam) and self.level < 3:
            self.update_subwrappers_tabular(
                self.parameter.params,
                self.delete_row_callback if not self.parameter.read_only else None,
            )
        elif isinstance(self.parameter, ArrayParam):
            self.update_subwrappers(
                self.parameter.params,
                self.delete_row_callback if not self.parameter.read_only else None,
            )
            if self.button:
                self.button["state"] = (
                    tk.DISABLED
                    if self.parameter.max_size is not None and len(self.subwrappers) >= self.parameter.max_size
                    else tk.NORMAL
                )
        elif isinstance(self.parameter, MultiParam):
            self.update_subwrappers(self.parameter.params.values(), None)
        elif isinstance(self.parameter, ChoiceParam):
            choices = self.parameter.choices or [""]
            self.entry["values"] = choices
        elif isinstance(self.parameter, BooleanParam):
            self.set_checkbox_value()
        elif isinstance(self.parameter, TextParam):
            if self.parameter.read_only:
                self.entry["state"] = "normal"
            self.entry.replace("1.0", tk.END, self.parameter.value)
            if self.parameter.read_only:
                self.entry["state"] = "disabled"
        elif isinstance(self.parameter, FileParam):
            self.entry["text"] = self.parameter.value
        else:
            self.var.set(self.parameter.value)

        if self.label:
            self.label["text"] = self.parameter.label

    def cull_subwrappers(self, params):
        params_set = set(params)
        for p, pw in self.subwrappers.items():
            if p not in params_set:
                pw.destroy()

    def update_subwrappers(self, params, delete_row_callback):
        for n, p in enumerate(params):
            if p in self.subwrappers:
                self.subwrappers[p].update()
            else:
                self.subwrappers[p] = ParameterWrapper(
                    self.entry,
                    p,
                    self.callback,
                    delete_row_callback,
                    level=self.level + 1,
                )
            self.subwrappers[p].set_row(n)

        self.cull_subwrappers(params)

    def update_subwrappers_tabular(self, params, delete_row_callback):
        while self.subwrapper_buttons:
            self.subwrapper_buttons.pop().destroy()
        while self.row_labels:
            self.row_labels.pop().destroy()
        while self.column_labels:
            self.column_labels.pop().destroy()

        for n, pp in enumerate(self.parameter.param.values()):
            column_label = tk.Label(self.entry, text=pp.label)
            self.column_labels.append(column_label)
            column_label.grid(row=0, column=n + 1, sticky=tk.EW, padx=10)
            self.entry.columnconfigure(n + 1, weight=1)

        for n, p in enumerate(params):
            row_label = tk.Label(self.entry, text=p.label)
            row_label.grid(row=n + 1, column=0, padx=10)
            self.row_labels.append(row_label)

            subparams = p.params.values()
            for m, pp in enumerate(subparams):
                if pp in self.subwrappers:
                    self.subwrappers[pp].update()
                else:
                    self.subwrappers[pp] = ParameterWrapper(
                        self.entry,
                        pp,
                        self.callback,
                        delete_row_callback,
                        level=self.level + 1,
                    )
                self.subwrappers[pp].entry.grid(row=n + 1, column=m + 1, padx=10)
            if not self.parameter.read_only and delete_row_callback:
                button = delete_button(self.entry, command=partial(delete_row_callback, self, n))
                button.grid(row=n + 1, column=len(subparams) + 1, padx=10)
                self.subwrapper_buttons.append(button)

        self.cull_subwrappers([pp for p in params for pp in p.params.values()])

    def update_subwrappers_framed(self, params, delete_row_callback):
        for n, p in enumerate(params):
            if p in self.subwrappers:
                self.subwrappers[p].update()
            else:
                label_frame_label = tk.Frame(self.entry)
                tk.Label(label_frame_label, text=p.label).grid(row=0, column=0, padx=10)
                label_frame = tk.LabelFrame(self.entry, labelwidget=label_frame_label, padx=10)
                label_frame.grid(sticky=tk.NSEW, padx=10)
                label_frame.columnconfigure(0, weight=1)

                def _command_drc(param, label_frame):
                    delete_row_callback(self.subwrappers[param])
                    label_frame.destroy()

                self.subwrappers[p] = ParameterWrapper(
                    label_frame,
                    p,
                    self.callback,
                    delete_row_callback,
                    level=self.level + 1,
                )
                if not self.parameter.read_only and delete_row_callback:
                    button = delete_button(label_frame_label, command=partial(_command_drc, p, label_frame))
                    button.grid(row=0, column=1, padx=10)

            self.subwrappers[p].entry.grid(row=n, column=0, padx=10)

        self.cull_subwrappers(params)

    def combobox_set(self, event):
        assert isinstance(self.entry, ttk.Combobox)
        if self.entry.current() != -1:
            self.entry.set(event.char)

    def set_row(self, row):
        if not self.label:
            self.entry.grid(row=row, column=0, columnspan=3)
        elif self.button:
            self.label.grid(row=row, column=0)
            self.button.grid(row=row, column=1)
            self.entry.grid(row=row, column=2)
        else:
            self.label.grid(row=row, column=0, columnspan=2)
            self.entry.grid(row=row, column=2)

    def add_row_callback(self, *_):
        assert isinstance(self.parameter, ArrayParam)

        if isinstance(self.parameter, FileSaveParam):
            file_types = self.parameter.file_types
            filename = filedialog.asksaveasfilename(filetypes=file_types)
            self.parameter.value = filename
        if isinstance(self.parameter, FileArrayParam):
            file_types = self.parameter.file_types
            filenames = filedialog.askopenfilenames(filetypes=file_types)
            self.parameter.add_files(filenames)
        else:
            self.parameter.add_row()

        if self.callback is not None:
            self.callback(self.parameter)

        self.update()

    def delete_row_callback(self, parameter_wrapper, row=None):
        assert isinstance(self.parameter, ArrayParam)

        if row is not None:
            self.parameter.del_row(row)
        else:
            self.parameter.del_subparam(parameter_wrapper.parameter)

        self.update()
        if self.callback is not None:
            self.callback(self.parameter)

    def change_file_callback(self, *_):
        file_types = self.parameter.file_types
        filename = filedialog.askopenfilename(filetypes=file_types)
        self.parameter.value = filename
        self.entry["text"] = self.parameter.value
        self.callback(self.parameter)

    def set_value(self, value):
        # self.parameter.value is a property, and some cleaning may occur, so we just
        # check before and after to see if it has changed.
        old_parameter_value = self.parameter.value
        self.parameter.value = value
        new_parameter_value = self.parameter.value
        if old_parameter_value != new_parameter_value and self.callback is not None:
            self.callback(self.parameter)
        return self.parameter.value

    def set_choice(self, choice):
        self.entry.current(choice)
        if self.parameter.choice != choice:
            self.parameter.choice = choice
            self.callback(self.parameter)

    def value_changed_callback(self, *_):
        if isinstance(self.parameter, ChoiceParam) and self.entry.current() != -1:
            self.set_choice(self.entry.current())
        else:
            self.var.set(self.set_value(self.var.get()))

    def widget_modified_callback(self, *_):
        # only gets called the *first* time a modification happens, unless
        # we reset the flag which says a change has happened ...
        value = self.entry.get(1.0, tk.END)
        self.entry.edit_modified(False)
        self.set_value(value)

    def set_checkbox_value(self):
        if self.parameter.hide:
            self.entry.set_value(None)
        elif self.parameter.value:
            self.entry.set_value(True)
        else:
            self.entry.set_value(False)

    def toggle_checkbox_callback(self, *_):
        if self.parameter.read_only or self.parameter.hide:
            # XXX warn?
            return
        value = not self.parameter.value
        self.parameter.value = value
        self.set_checkbox_value()
        if self.callback is not None:
            self.callback(self.parameter)

    def clear_value_callback(self, *_):
        self.set_value("")

    def destroy(self):
        self.label.destroy()
        self.entry.destroy()
        if self.button:
            self.button.destroy()


class PluginConfigurator:
    output_text = None
    preview = None

    def __init__(self, tk_parent: tk.Widget, plugin: BasePlugin, change_callback=None):
        self.plugin = plugin
        self.change_callback = change_callback

        self.frame = tk.Frame(tk_parent)
        self.frame.columnconfigure(0, weight=1)
        self.frame.grid(sticky=tk.NSEW)

        self.wrapper_cache: MutableMapping[str, ParameterWrapper] = {}

        self.subframe = tk.Frame(self.frame)
        self.subframe.columnconfigure(0, weight=0)
        self.subframe.columnconfigure(1, weight=0)
        self.subframe.columnconfigure(2, weight=1)
        self.subframe.grid(row=1, sticky=tk.NSEW)

        self.update()

    def change_parameter(self, parameter):
        """Called whenever a parameter gets changed"""
        # if self.plugin.update():
        #    self.update()
        if self.change_callback:
            self.change_callback(self)

    def update(self) -> None:
        # If there's only a single parameter it is presented a little differently.
        top_level = 0 if len(self.plugin.parameters) == 1 else 1

        # Create any new parameter wrappers needed & update existing ones
        for n, (key, parameter) in enumerate(self.plugin.parameters.items()):
            if key in self.wrapper_cache:
                self.wrapper_cache[key].update()
            else:
                self.wrapper_cache[key] = ParameterWrapper(
                    self.subframe, parameter, self.change_parameter, level=top_level
                )
            self.wrapper_cache[key].set_row(n + 1)

        # Remove any parameter wrappers no longer needed
        for key, wrapper in list(self.wrapper_cache.items()):
            if key not in self.plugin.parameters:
                wrapper.destroy()
                del self.wrapper_cache[key]



import tkinter as tk
import webbrowser
from urllib.parse import urlparse

from tkinterweb import HtmlFrame  # type: ignore

MINI_CSS = """
    * { padding: 10px; line-height: 150% }
    th, td { border: 1px solid #AAA; border-collapse: collapse; }
    code { border: 1px solid #AAA; font-family: monospace; }
    th code, td code { border: 0px; }
"""


class MiniBrowserFrame(tk.Frame):
    def __init__(self, tk_parent, start_url, *a, **k):
        super().__init__(tk_parent, *a, **k)
        tk.Label(self, text="Documentation Preview").pack(fill="both")

        self.html_frame = HtmlFrame(self, messages_enabled=False)
        self.html_frame.enable_stylesheets(False)
        self.html_frame.enable_objects(False)
        self.html_frame.enable_forms(False)
        self.html_frame.on_done_loading(self.on_done_loading)
        self.html_frame.on_link_click(self.on_link_click)

        self.html_frame.pack(fill="both", expand=True)

        tk.Button(self, text="Open in Browser", command=self.on_browser_button).pack()

        self.load_url(start_url)

    def load_url(self, link_url):
        self.current_url = link_url
        self.html_frame.load_url(link_url)

    def on_browser_button(self):
        webbrowser.open_new_tab(self.current_url)

    def on_done_loading(self):
        self.html_frame.add_css(MINI_CSS)

    def on_link_click(self, link_url):
        if urlparse(self.current_url)[0:2] == urlparse(link_url)[0:2]:
            self.load_url(link_url)
        else:
            webbrowser.open_new_tab(link_url)


if __name__ == "__main__":
    root = tk.Tk()
    url = "https://countess-project.github.io/CountESS/"
    MiniBrowserFrame(root, url).pack(fill="both", expand=True)
    root.mainloop()



from typing import Iterable, Optional

import pandas as pd

from countess import VERSION
from countess.core.logger import Logger
from countess.core.parameters import (
    ArrayParam,
    BooleanParam,
    DataTypeChoiceParam,
    MultiParam,
    StringParam,
    TabularMultiParam,
)
from countess.core.plugins import PandasInputPlugin


class DataTablePlugin(PandasInputPlugin):
    """DataTable"""

    name = "DataTable"
    description = "enter small amounts of data directly"
    link = "https://countess-project.github.io/CountESS/included-plugins/#data-table"
    version = VERSION

    parameters = {
        "columns": ArrayParam(
            "Columns",
            MultiParam(
                "Column",
                {
                    "name": StringParam("Name"),
                    "type": DataTypeChoiceParam("Type", "string"),
                    "index": BooleanParam("Index?"),
                },
            ),
        ),
        "rows": ArrayParam("Rows", TabularMultiParam("Row", {})),
    }

    col_rows: list[MultiParam] = []

    show_preview = False

    def fix_columns(self):
        for num, col_row in enumerate(self.col_rows):
            if col_row not in self.parameters["columns"].params:
                row_params = self.parameters["rows"].param.params
                # XXX a column has been deleted!
                for col_num in range(num, len(self.col_rows) - 1):
                    row_params[str(col_num)] = row_params[str(col_num + 1)]
                    for row in self.parameters["rows"]:
                        row.params[str(col_num)] = row.params[str(col_num + 1)]
                del row_params[str(len(self.col_rows) - 1)]
                for row in self.parameters["rows"]:
                    del row.params[str(len(self.col_rows) - 1)]
        self.col_rows = self.parameters["columns"].params[:]

        for num, col in enumerate(self.parameters["columns"]):
            col_name = col["name"].value
            try:
                self.parameters["rows"].param.params[str(num)].label = col_name
            except KeyError:
                new_param = col["type"].get_parameter(col_name)
                self.parameters["rows"].param.params[str(num)] = new_param
            for row in self.parameters["rows"]:
                try:
                    new_param = col["type"].get_parameter(col_name)
                    cur_param = row.params[str(num)]
                    if type(cur_param) != type(new_param):
                        try:
                            new_param.value = cur_param.value
                        except ValueError:
                            pass
                        row.params[str(num)] = new_param
                    row.params[str(num)].label = col_name
                except KeyError:
                    p = col["type"].get_parameter(col_name)
                    row.params[str(num)] = p

    def set_parameter(self, key: str, *a, **k):
        if key == "rows.0.0":
            self.fix_columns()
        super().set_parameter(key, *a, **k)

    def num_files(self):
        return 1

    def load_file(self, file_number: int, logger: Logger, row_limit: Optional[int] = None) -> Iterable[pd.DataFrame]:
        assert file_number == 0
        assert isinstance(self.parameters["rows"], ArrayParam)
        assert isinstance(self.parameters["columns"], ArrayParam)
        self.fix_columns()
        values = []
        for row in self.parameters["rows"]:
            values.append(
                dict((col["name"].value, row[str(num)].value) for num, col in enumerate(self.parameters["columns"]))
            )

        df = pd.DataFrame(values)

        index_cols = [col["name"].value for col in self.parameters["columns"] if col["index"].value]
        if index_cols:
            df = df.set_index(index_cols)

        yield df



from typing import Optional

from fqfa.util.nucleotide import reverse_complement  # type: ignore

from countess import VERSION
from countess.core.logger import Logger
from countess.core.parameters import BooleanParam, ColumnChoiceParam, IntegerParam, StringParam
from countess.core.plugins import PandasTransformSingleToSinglePlugin


class SequencePlugin(PandasTransformSingleToSinglePlugin):
    """Manipulate DNA valueuences"""

    name = "Sequence Tool"
    description = "Manipulate DNA Sequences"
    version = VERSION
    link = "https://countess-project.github.io/CountESS/included-plugins/#valueuence"

    parameters = {
        "column": ColumnChoiceParam("Input Column"),
        "invert": BooleanParam("Invert", False),
        "offset": IntegerParam("Offset", 0),
        "start": StringParam("Start at ...", ""),
        "stop": StringParam("Stop at ...", ""),
        "length": IntegerParam("Max Length", 150),
        "output": StringParam("Output Column", "sequence"),
    }

    def process_value(self, value: str, logger: Logger) -> Optional[str]:
        if value is None:
            return None

        if self.parameters["invert"].value:
            value = reverse_complement(value)
        if self.parameters["offset"].value:
            offset = self.parameters["offset"].value
            value = value[offset:]
        if self.parameters["start"].value:
            offset = value.find(self.parameters["start"].value)
            if offset >= 0:
                value = value[offset:]
            else:
                return None
        if self.parameters["stop"].value:
            offset = value.find(self.parameters["stop"].value)
            if offset >= 0:
                value = value[0 : offset + len(self.parameters["stop"].value)]
        if self.parameters["length"].value:
            value = value[0 : self.parameters["length"].value]
        return value



from typing import Iterable, List

import numpy as np
import pandas as pd
from pandas.api.typing import DataFrameGroupBy  # type: ignore

from countess import VERSION
from countess.core.logger import Logger
from countess.core.parameters import ArrayParam, BooleanParam, PerColumnArrayParam, TabularMultiParam
from countess.core.plugins import PandasProcessPlugin
from countess.utils.pandas import flatten_columns, get_all_columns


class GroupByPlugin(PandasProcessPlugin):
    """Groups a Pandas Dataframe by an arbitrary column and rolls up rows"""

    name = "Group By"
    description = "Group records by column(s) and calculate aggregates"
    version = VERSION
    link = "https://countess-project.github.io/CountESS/included-plugins/#group-by"

    input_columns: dict[str, np.dtype]

    parameters = {
        "columns": PerColumnArrayParam(
            "Columns",
            TabularMultiParam(
                "Column",
                {
                    "index": BooleanParam("Index"),
                    "count": BooleanParam("Count"),
                    "min": BooleanParam("Min"),
                    "max": BooleanParam("Max"),
                    "sum": BooleanParam("Sum"),
                    "mean": BooleanParam("Mean"),
                },
            ),
        ),
        "join": BooleanParam("Join Back?"),
    }

    dataframes: List[pd.DataFrame]

    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self.prepare()

    def prepare(self, *_):
        self.dataframes = []
        self.input_columns = {}

    def process(self, data: pd.DataFrame, source: str, logger: Logger) -> Iterable:
        # XXX should do this in two stages: group each dataframe and then combine.
        # that can wait for a more general MapReduceFinalizePlugin class though.
        assert self.dataframes is not None
        assert isinstance(self.parameters["columns"], ArrayParam)

        self.input_columns.update(get_all_columns(data))
        column_parameters = list(zip(self.input_columns.keys(), self.parameters["columns"].params))

        if not self.parameters["join"].value:
            # Dispose of any columns we don't use in the aggregations.
            # TODO: Reindex as well?
            keep_columns = [
                col
                for col, col_param in column_parameters
                if isinstance(col_param, TabularMultiParam)
                and any(cp.value for cp in col_param.values())
                and col in data.columns
            ]
            data = data[keep_columns]

        self.dataframes.append(data)
        return []

    def finalize(self, logger: Logger) -> Iterable[pd.DataFrame]:
        assert isinstance(self.parameters["columns"], ArrayParam)
        assert self.dataframes

        self.parameters["columns"].set_column_choices(self.input_columns.keys())

        column_parameters = list(zip(self.input_columns.keys(), self.parameters["columns"]))

        index_cols = [col for col, col_param in column_parameters if col_param["index"].value]
        agg_ops = dict(
            (
                col,
                [k for k, pp in col_param.params.items() if pp.value and k != "index"],
            )
            for col, col_param in column_parameters
            if any(pp.value for k, pp in col_param.params.items() if k != "index")
        )

        data_in = pd.concat(self.dataframes)
        data_in.reset_index([col for col in agg_ops.keys() if col in data_in.index.names], inplace=True)

        try:
            # If there are no columns to index by, add a dummy column and group by that so we
            # still get a DataFrameGroupBy for the next operation
            data_grouped: DataFrameGroupBy = (
                data_in.groupby(index_cols) if index_cols else data_in.assign(__temp=1).groupby("__temp")
            )

            # If no aggregation operations have been selected then just count the groups.
            if agg_ops:
                data_out = data_grouped.agg(agg_ops)
            else:
                data_out = pd.DataFrame(data_grouped.size(), columns=["count"])

            # Get rid of that dummy column if we added it.
            if "__temp" in data_out.index.names:
                data_out.reset_index("__temp", drop=True, inplace=True)

            flatten_columns(data_out, inplace=True)

            if self.parameters["join"].value:
                if index_cols:
                    yield data_in.merge(data_out, how="left", left_on=index_cols, right_on=index_cols)
                else:
                    yield data_in.assign(**data_out.to_dict("records")[0])
            else:
                yield data_out
        except ValueError as exc:
            logger.exception(exc)



import pandas as pd

from countess import VERSION
from countess.core.logger import Logger
from countess.core.parameters import BooleanParam, PerColumnArrayParam, TextParam
from countess.core.plugins import PandasSimplePlugin


def process(df: pd.DataFrame, codes, logger: Logger):
    for code in codes:
        if not code:
            continue

        try:
            result = df.eval(code)
        except Exception as exc:  # pylint: disable=W0718
            logger.exception(exc)
            continue

        if isinstance(result, pd.Series):
            # this was a filter
            df = df.copy()
            df["__filter"] = result
            df = df.query("__filter != 0").drop(columns="__filter")
        else:
            # this was a column assignment
            df = result

    return df


class ExpressionPlugin(PandasSimplePlugin):
    name = "Expression"
    description = "Apply simple expressions"
    version = VERSION

    parameters = {
        "code": TextParam("Expressions"),
        "drop": PerColumnArrayParam("Drop Columns", BooleanParam("Drop")),
    }

    def process_dataframe(self, dataframe: pd.DataFrame, logger: Logger) -> pd.DataFrame:
        assert isinstance(self.parameters["drop"], PerColumnArrayParam)

        codes = [c.replace("\n", " ").strip() for c in self.parameters["code"].value.split("\n\n")]
        df = process(dataframe, codes, logger)

        drop_names = [label for label, param in self.parameters["drop"].get_column_params() if param.value]

        drop_indexes = [col for col in drop_names if col in df.index.names]
        if drop_indexes:
            df = df.reset_index(drop_indexes, drop=True)

        drop_columns = [col for col in drop_names if col in df.columns]
        if drop_columns:
            df.drop(columns=drop_columns, inplace=True)

        return df



import pandas as pd

from countess import VERSION
from countess.core.logger import Logger
from countess.core.parameters import (
    ArrayParam,
    BooleanParam,
    DataTypeOrNoneChoiceParam,
    PerColumnArrayParam,
    StringParam,
    TabularMultiParam,
)
from countess.core.plugins import PandasSimplePlugin


class ColumnToolPlugin(PandasSimplePlugin):
    name = "DataFrame Column Tool"
    description = "Alter Columns of a DataFrame"
    version = VERSION

    parameters = {
        "columns": PerColumnArrayParam(
            "Columns",
            TabularMultiParam(
                "Column",
                {
                    "rename": StringParam("Name"),
                    "datatype": DataTypeOrNoneChoiceParam("Column Type"),
                    "index": BooleanParam("Index?"),
                },
            ),
        )
    }

    def process_dataframe(self, dataframe: pd.DataFrame, logger: Logger) -> pd.DataFrame:
        assert isinstance(self.parameters["columns"], ArrayParam)
        column_parameters = list(zip(self.input_columns, self.parameters["columns"]))

        drop_columns = [column_name for column_name, parameter in column_parameters if parameter.datatype.is_none()]

        type_columns = {
            column_name: parameter.datatype.get_selected_type()
            for column_name, parameter in column_parameters
            if parameter.datatype.is_not_none()
        }

        index_columns = [
            column_name
            for column_name, parameter in column_parameters
            if parameter.index.value and parameter.datatype.is_not_none()
        ]

        rename_columns = {
            column_name: parameter.rename.value.strip()
            for column_name, parameter in column_parameters
            if parameter.rename.value and parameter.rename.value.strip() and parameter.datatype.is_not_none()
        }

        dataframe = dataframe.reset_index().drop(columns=drop_columns).astype(type_columns)

        if index_columns:
            dataframe = dataframe.set_index(index_columns)

        return dataframe.rename(columns=rename_columns)



import gc
import threading
import time
from multiprocessing import Process, Queue, Value
from os import cpu_count, getpid
from queue import Empty
from typing import Callable, Iterable

try:
    from typing import Concatenate, ParamSpec, TypeVar
except ImportError:
    # for Python 3.9 compatibility
    from typing_extensions import Concatenate, ParamSpec, TypeVar  # type: ignore

import psutil

D = TypeVar("D")
V = TypeVar("V")
P = ParamSpec("P")


def multiprocess_map(
    function: Callable[Concatenate[V, P], Iterable[D]], values: Iterable[V], *args: P.args, **kwargs: P.kwargs
) -> Iterable[D]:
    """Pretty much equivalent to:

        def multiprocess_map(function, values, *args, **kwargs):
            yield from interleave_longest(function(v, *args, **kwargs) for v in values)

    but runs in multiple processes using an input_queue
    to organize `values` and an output_queue to organize the
    returned values.  Or to look at it from a different direction, because
    `multiprocessing.Pool.imap_unordered` doesn't take `args`."""

    # Start up several workers.
    nproc = ((cpu_count() or 1) + 1) // 2
    input_queue: Queue = Queue()
    output_queue: Queue = Queue(maxsize=3)

    # XXX is it actually necessary to have this in a separate thread or
    # would it be sufficient to add items to input_queue alternately with
    # removing items from output_queue?

    enqueue_running = Value("b", True)

    def __enqueue():
        for v in values:
            input_queue.put(v)
        enqueue_running.value = False

    thread = threading.Thread(target=__enqueue)
    thread.start()

    # XXX is this necessary?
    time.sleep(1)

    def __process():
        while True:
            try:
                while True:
                    # Prevent processes from using up all
                    # available memory while waiting
                    # XXX this is probably a bad idea
                    while psutil.virtual_memory().percent > 75:
                        print(f"{getpid()} LOW MEMORY {psutil.virtual_memory().percent}")
                        time.sleep(1)

                    data_in = input_queue.get(timeout=1)
                    for data_out in function(data_in, *args, **(kwargs or {})):
                        output_queue.put(data_out)

                        # Make sure large data is disposed of before we
                        # go around for the next loop
                        del data_out
                    del data_in
                    gc.collect()

            except Empty:
                if not enqueue_running.value:
                    break

    processes = [Process(target=__process, name=f"worker {n}") for n in range(0, nproc)]
    for p in processes:
        p.start()

    while thread.is_alive() or any(p.is_alive() for p in processes):
        try:
            yield output_queue.get(timeout=0.1)
        except Empty:
            # Waiting for the next input, might as well tidy up
            gc.collect()



import datetime
import tkinter as tk
from tkinter import ttk

from countess.core.logger import MultiprocessLogger


class LoggerTreeview(ttk.Treeview):
    def __init__(self, tk_parent, *a, **k):
        super().__init__(tk_parent, *a, **k)
        self["columns"] = ["name", "message"]
        self.heading(0, text="name")
        self.heading(1, text="message")

    def clear(self):
        self.delete(*self.get_children())


class LabeledProgressbar(ttk.Progressbar):
    """A progress bar with a label on top of it, the progress bar value can be set in the
    usual way and the label can be set with self.update_label"""

    # see https://stackoverflow.com/a/40348163/90927 for how the styling works.

    style_data = [
        (
            "LabeledProgressbar.trough",
            {
                "children": [
                    ("LabeledProgressbar.pbar", {"side": "left", "sticky": tk.NS}),
                    ("LabeledProgressbar.label", {"sticky": ""}),
                ],
                "sticky": tk.NSEW,
            },
        )
    ]

    def __init__(self, master, *args, **kwargs):
        self.style = ttk.Style(master)
        # make up a new style name so we don't interfere with other LabeledProgressbars
        # and accidentally change their color or label (uses arbitrary object ID)
        self.style_name = f"_id_{id(self)}"
        self.style.layout(self.style_name, self.style_data)
        self.style.configure(self.style_name, background="green")

        kwargs["style"] = self.style_name
        super().__init__(master, *args, **kwargs)

    def update_label(self, s):
        self.style.configure(self.style_name, text=s)


class LoggerFrame(ttk.Frame):
    def __init__(self, tk_parent, *a, **k):
        super().__init__(tk_parent, *a, **k)
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)

        self.treeview = LoggerTreeview(self)
        self.treeview.grid(row=0, column=0, sticky=tk.NSEW)
        self.treeview.bind("<<TreeviewSelect>>", self.on_click)

        self.scrollbar_x = ttk.Scrollbar(self, orient=tk.HORIZONTAL, command=self.treeview.xview)
        self.scrollbar_x.grid(row=1, column=0, sticky=tk.EW)
        self.treeview.configure(xscrollcommand=self.scrollbar_x.set)

        self.scrollbar_y = ttk.Scrollbar(self, orient=tk.VERTICAL, command=self.treeview.yview)
        self.scrollbar_y.grid(row=0, column=1, sticky=tk.NS)
        self.treeview.configure(yscrollcommand=self.scrollbar_y.set)

        self.progress_frame = tk.Frame(self)
        self.progress_frame.grid(row=2, columnspan=2, sticky=tk.EW)
        self.progress_frame.columnconfigure(0, weight=1)

        self.logger = MultiprocessLogger()
        self.count = 0
        self.details = {}
        self.progress_bars = {}

        self.poll()

    def poll(self):
        datetime_now = datetime.datetime.now()
        for level, message, detail in self.logger.poll():
            if level == "progress":
                try:
                    pbar = self.progress_bars[message]
                except KeyError:
                    pbar = LabeledProgressbar(self.progress_frame, mode="determinate", value=0)
                    self.progress_bars[message] = pbar
                    pbar.update_label(message)
                    pbar.grid(sticky=tk.EW)

                if detail is not None:
                    progress = int(float(detail))
                    pbar.config(mode="determinate", value=progress)
                    pbar.update_label(f"{message} {detail}%")
                    if progress == 100:
                        self.after(2000, self.remove_pbar, message)
                else:
                    pbar.config(mode="indeterminate")
                    pbar.step(5)
                    pbar.update_label(f"{message}")

            else:
                self.count += 1
                iid = self.treeview.insert("", "end", text=datetime_now.isoformat(), values=(level, message))
                if detail:
                    self.details[iid] = detail
        self.after(100, self.poll)

    def get_logger(self, name: str):
        return self.logger

    def on_click(self, _):
        # XXX display detail more nicely
        TreeviewDetailWindow(self.details[self.treeview.focus()])

    def remove_pbar(self, message):
        if message in self.progress_bars:
            self.progress_bars[message].destroy()
            del self.progress_bars[message]

    def clear(self):
        self.logger.clear()
        self.treeview.clear()
        self.count = 0
        self.details = {}
        for message in self.progress_bars:
            self.remove_pbar(message)


class TreeviewDetailWindow(tk.Toplevel):
    def __init__(self, detail, *a, **k):
        super().__init__(*a, **k)

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

        text = tk.Text(self)
        text.insert("1.0", detail)
        text["state"] = "disabled"
        text.grid(sticky=tk.NSEW)

        button = tk.Button(self, text="CLOSE", command=self.destroy)
        button.grid(sticky=tk.EW)



""" pandas utility functions """

from typing import Any, Dict, Iterable

import pandas as pd


def collect_dataframes(data: Iterable[pd.DataFrame], preferred_size: int = 100000) -> Iterable[pd.DataFrame]:
    """Takes an iterable of dataframes, ignores any empty ones and tries to pack smaller ones
    together up to `preferred_size`."""
    # XXX Data doesn't necessarily come out in the same order it went in.
    buffer = None
    for dataframe in data:
        if dataframe is None or len(dataframe) == 0:
            continue
        if len(dataframe) > preferred_size:
            yield dataframe
        elif buffer is None:
            buffer = dataframe
        elif len(buffer) + len(dataframe) > preferred_size:
            yield buffer
            buffer = dataframe
        else:
            # XXX catch errors?
            buffer = pd.concat([buffer, dataframe])
    if buffer is not None and len(buffer) > 0:
        yield buffer


def get_all_indexes(dataframe: pd.DataFrame) -> Dict[str, Any]:
    """Indexes work a couple of different ways, so this massages them into a consistent
    dictionary of name => dtype.  A single RangeIndex or a nameless index isn't counted."""
    if isinstance(dataframe.index, pd.RangeIndex):
        return {}
    elif dataframe.index.name:
        return {str(dataframe.index.name): dataframe.index.dtype}
    elif (
        hasattr(dataframe.index, "names")
        and hasattr(dataframe.index, "dtypes")
        and dataframe.index.names[0] is not None
    ):
        return dict(zip(dataframe.index.names, dataframe.index.dtypes))
    else:
        # XXX I'm not sure how or if this ever happens
        return {}  # pragma: no cover


def get_all_columns(dataframe: pd.DataFrame) -> Dict[str, Any]:
    r = get_all_indexes(dataframe)
    r.update(dict(zip(dataframe.columns, dataframe.dtypes)))
    return r


def concat_dataframes(dataframes: list[pd.DataFrame]) -> pd.DataFrame:
    """handles the special case where there's an empty list"""
    # XXX dask also had a special case for a single input dataframe I think?
    if len(dataframes) == 0:
        return pd.DataFrame()
    return pd.concat(dataframes)


def flatten_columns(dataframe: pd.DataFrame, inplace: bool = False) -> pd.DataFrame:
    """used to rename columns with tuple names like `('things', 'count')` to easier-to-refer-to-
    in-python names like `'things__count'`."""
    # XXX arguably this would be better done at python plugin and/or export time but for now
    # we flatten columns where possible.

    if not inplace:
        dataframe = dataframe.copy()

    dataframe.columns = [
        "__".join(a) if type(a) is tuple else a for a in dataframe.columns.to_flat_index()
    ]  # type: ignore [assignment]

    return dataframe



import io
import tkinter as tk
from functools import partial
from math import ceil, floor, isinf, isnan
from tkinter import ttk
from typing import Callable, Optional, Union

import pandas as pd
from pandas.api.types import is_integer_dtype, is_numeric_dtype

from countess.gui.widgets import get_icon

# XXX columns should automatically resize based on information
# from _column_xscrollcommand which can tell if they're
# overflowing.  Or maybe use
#    df['seq'].str.len().max() to get max length of a string column
# etc etc.


def column_format_for(df_column: Union[pd.Index, pd.Series]) -> str:
    if is_numeric_dtype(df_column.dtype):
        # Work out the maximum width required to represent the integer part in this
        # column, so we can pad values to that width.
        column_min = df_column.min()
        if isnan(column_min) or isinf(column_min):
            column_min = -100
        column_max = df_column.max()
        if isnan(column_max) or isinf(column_max):
            column_max = 100
        width = max(len(str(floor(column_min))), len(str(ceil(column_max))))
        if is_integer_dtype(df_column.dtype):
            return f"%{width}d"
        else:
            # leave room for the point and 12 decimals.
            # format_value will remove trailing 0s.
            return f"%{width+13}.12f"
    else:
        return "%s"


def format_value(value: Optional[Union[int, float, str]], column_format: str) -> str:
    """Format value for display in a table:
    >>> format_value(None, "%s")
    '—'
    >>> format_value(107, "%4d")
    ' 107'
    >>> format_value(1.23, "%15.12f")
    ' 1.23'
    >>> format_value("foo", "%15.12f")
    'foo'

    """

    if value is None or (type(value) is float and isnan(value)):
        return "—"
    if value is True:
        return "—T"
    if value is False:
        return "—F"

    # remove trailing 0's from floats (%g doesn't align correctly)
    try:
        if column_format.endswith("f"):
            return (column_format % value).rstrip("0")
        else:
            return column_format % value
    except TypeError:
        return str(value)


class TabularDataFrame(tk.Frame):
    """A frame for displaying a pandas (or similar) dataframe.
    Columns are displayed as individual tk.Text widgets which seems to be relatively efficient
    as they only hold the currently displayed rows.  Tested up to a million or so rows."""

    subframe: Optional[tk.Frame] = None
    dataframe: Optional[pd.DataFrame] = None
    offset = 0
    height = 1000
    length = 0
    select_rows = None
    labels: list[tk.Label] = []
    columns: list[tk.Text] = []
    column_formats: list[str] = []
    scrollbar = None
    index_cols = 0
    sort_by_col = None
    sort_ascending = True
    callback: Optional[Callable[[int, int, bool], None]] = None

    def reset(self):
        if self.subframe:
            self.subframe.destroy()
        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)
        self.subframe = tk.Frame(self)
        self.subframe.rowconfigure(0, weight=0)
        self.subframe.rowconfigure(1, weight=0)
        self.subframe.rowconfigure(2, weight=1)
        self.subframe.grid(sticky=tk.NSEW)

    def set_dataframe(self, dataframe: pd.DataFrame, offset: Optional[int] = 0):
        self.reset()
        assert self.subframe

        self.dataframe = dataframe
        self.length = len(dataframe)

        if hasattr(self.dataframe.index, "names") and hasattr(self.dataframe.index, "dtypes"):
            # MultiIndex case
            column_names = list(self.dataframe.index.names) + list(self.dataframe.columns)
            column_dtypes = list(self.dataframe.index.dtypes) + list(self.dataframe.dtypes)
            index_frame = self.dataframe.index.to_frame()
            self.column_formats = [column_format_for(index_frame[name]) for name in dataframe.index.names] + [
                column_format_for(dataframe[name]) for name in dataframe.columns
            ]
            self.index_cols = len(self.dataframe.index.names)
        elif self.dataframe.index.name:
            # a simple Index, with a name
            column_names = [self.dataframe.index.name] + list(self.dataframe.columns)
            column_dtypes = [self.dataframe.index.dtype] + list(self.dataframe.dtypes)
            self.column_formats = [column_format_for(dataframe.index)] + [
                column_format_for(dataframe[name]) for name in dataframe.columns
            ]
            self.index_cols = 1
        else:
            # if it doesn't have a name, don't bother displaying it
            # XXX it's probably just a RangeIndex, should we display it anyway?
            column_names = list(self.dataframe.columns)
            column_dtypes = list(self.dataframe.dtypes)
            self.column_formats = [column_format_for(dataframe[name]) for name in dataframe.columns]
            self.index_cols = 0

        if len(column_names) == 0:
            label = tk.Label(self.subframe, text="Dataframe Preview\n\nno data")
            label.grid(row=0, column=0, sticky=tk.NSEW)
            return

        title = tk.Label(self.subframe, text=f"Dataframe Preview {len(self.dataframe)} rows")
        title.grid(row=0, column=0, columnspan=len(column_names) * 2 + 1, sticky=tk.NSEW, pady=5)

        ### XXX add in proper handling for MultiIndexes here

        # Even-numbered columns are the data columns

        self.labels = []
        for num, (name, dtype) in enumerate(zip(column_names, column_dtypes)):
            if type(name) is tuple:
                name = "\n".join([str(n) for n in name])
            else:
                name = str(name)
            is_index = " (index)" if num < self.index_cols else ""
            label = tk.Label(
                self.subframe,
                text=f"{name}\n{dtype}{is_index}",
                image=get_icon(self, "sort_un"),
                compound=tk.RIGHT,
            )
            label.grid(row=1, column=num * 2, sticky=tk.EW)
            label.bind("<Button-1>", partial(self._label_button_1, num))
            self.subframe.columnconfigure(num * 2, minsize=10, weight=1)
            self.labels.append(label)

        # Between them are blank columns which provide a handle for adjusting the column
        # widths left and right

        for num in range(0, len(column_names) - 1):
            adjuster = tk.Frame(self.subframe, width=3, cursor="sb_h_double_arrow")
            adjuster.grid(row=1, rowspan=2, column=num * 2 + 1, sticky=tk.NSEW)
            adjuster.bind("<B1-Motion>", partial(self._column_adjust, num))

        if len(self.dataframe) == 0:
            label = tk.Label(self.subframe, text="no data")
            label.grid(row=2, column=0, columnspan=len(column_names), sticky=tk.NSEW)
            return

        self.columns = [tk.Text(self.subframe) for _ in column_names]
        for num, column in enumerate(self.columns):
            column.grid(sticky=tk.NSEW, row=2, column=num * 2)
            column["wrap"] = tk.NONE
            column["xscrollcommand"] = partial(self._column_xscrollcommand, num)
            column["yscrollcommand"] = self._column_yscrollcommand
            column.bind("<Button-4>", self._column_scroll)
            column.bind("<Button-5>", self._column_scroll)
            column.bind("<<Selection>>", partial(self._column_selection, num))
            column.bind("<Control-C>", self._column_copy)
            column.bind("<<Copy>>", self._column_copy)
        if self.columns:
            self.columns[0].bind("<Configure>", self._column_configure)

        self.scrollbar = ttk.Scrollbar(self.subframe, orient=tk.VERTICAL)
        self.scrollbar.grid(sticky=tk.NS, row=2, column=len(self.columns) * 2 - 1)
        self.scrollbar["command"] = self._scrollbar_command
        self.refresh(offset)

    def refresh(self, new_offset=0):
        # Refreshes the column widgets.
        # XXX should handle new_height as well, as this changes a fair bit
        # with some window managers. Needs refactoring.

        new_offset = max(0, min(self.length - self.height, int(new_offset)))
        offset_diff = new_offset - self.offset

        # get the new rows as an iterator
        if 1 <= offset_diff < self.height:
            df = self.dataframe.iloc[self.offset + self.height : self.offset + self.height + offset_diff]
            insert_at = tk.END
        elif 1 <= -offset_diff < self.height:
            # Get rows in reverse order so they can be inserted at the start
            # note offset_diff is negative!
            # XXX check this isn't horribly inefficient with pandas indexes
            if new_offset:
                df = self.dataframe.iloc[self.offset - 1 : new_offset - 1 : -1]
            else:
                df = self.dataframe.iloc[self.offset - 1 :: -1]
            insert_at = "1.0"
        else:
            df = self.dataframe.iloc[new_offset : new_offset + self.height + 1]
            insert_at = tk.END

        # then unlock the columns and delete unnecessary rows
        for cw in self.columns:
            cw["state"] = tk.NORMAL
            if 1 <= offset_diff < self.height:
                # delete rows at start, add new rows on the end
                cw.delete("1.0", f"{offset_diff+1}.0")
            elif 1 <= -offset_diff < self.height:
                # delete rows at end, insert new rows at start
                # note offset_diff is negative!  Note we have to
                # restore the deleted final "\n".  Sigh.
                cw.delete(f"{self.height+offset_diff+2}.0", tk.END)
                cw.insert(tk.END, "\n")
            else:
                # delete everything & add all new rows
                cw.delete("1.0", tk.END)

        for idx, row in df.iterrows():
            if type(idx) is tuple:
                values = list(idx) + list(row)
            elif self.dataframe.index.name:
                values = [idx] + list(row)
            else:
                values = list(row)

            for value, column, column_format in zip(values, self.columns, self.column_formats):
                column.insert(insert_at, format_value(value, column_format) + "\n")

        for cw in self.columns:
            cw["state"] = tk.DISABLED

        self.offset = new_offset
        if self.length:
            self.scrollbar.set(self.offset / self.length, (self.offset + self.height) / self.length)

        if self.callback:
            self.callback(self.offset, self.sort_by_col, not self.sort_ascending)

    def set_callback(self, callback) -> None:
        self.callback = callback

    def set_sort_order(self, column_num: int, descending: Optional[bool] = None):
        assert self.dataframe is not None

        if descending is None and column_num == self.sort_by_col:
            self.sort_ascending = not self.sort_ascending
        else:
            self.sort_by_col = column_num
            self.sort_ascending = not descending
        if column_num < self.index_cols:
            self.dataframe = self.dataframe.sort_index(level=column_num, ascending=self.sort_ascending)
        else:
            self.dataframe = self.dataframe.sort_values(
                self.dataframe.columns[column_num - self.index_cols], ascending=self.sort_ascending
            )

        for n, label in enumerate(self.labels):
            icon = "sort_un" if n != column_num else "sort_up" if self.sort_ascending else "sort_dn"
            label.configure(image=get_icon(self, icon))

        self.refresh()

    def _label_button_1(self, num, event):
        """Click on column labels to set sort order"""
        self.set_sort_order(num)
        if self.callback:
            self.callback(self.offset, self.sort_by_col, not self.sort_ascending)

    def _column_adjust(self, num, event):
        """Adjust column widths left and right by dragging the dummy columns"""
        w0 = self.labels[num].winfo_width()
        w1 = self.labels[num + 1].winfo_width()
        self.subframe.columnconfigure(num * 2, minsize=w0 + event.x)
        self.subframe.columnconfigure(num * 2 + 2, minsize=w1 - event.x)

    def _scrollbar_command(self, command, *parameters):
        # Detect scrollbar movement and move self.offset
        # to compensate.
        if command == "moveto":
            self.refresh(float(parameters[0]) * (self.length - self.height))
        elif command == "scroll":
            self.refresh(self.offset + int(parameters[0]))
        else:
            self.refresh()

    def _column_xscrollcommand(self, num, x1, x2):
        # XXX this gets called as the table is displayed
        # if x2 < 1.0 this column is partially hidden.
        pass

    # XXX the following two functions are clever but not
    # very efficient!  There's got to be a nicer way of
    # detecting this surely?

    def _column_yscrollcommand(self, y1, y2):
        # All this actually does is to detect if there's
        # too many rows for the window, in which case it
        # trims them off.  Once there's the right number
        # of rows, this won't get called any more.
        span = int((float(y2) - float(y1)) * self.height)
        if span > 0 and span != self.height:
            self.height = span
            self.refresh()

    def _column_configure(self, *_):
        # If we've resized the window, start with a huge
        # number of rows and let _cw_yscrollcommand trim
        # it back down again.  Probably could be more
        # sensible.
        self.height = min(self.length, 1000)
        self.refresh()

    def _column_scroll(self, event):
        # Detect scrollwheel motion on any of the columns
        if event.num == 4:
            self.refresh(self.offset - 5)
        elif event.num == 5:
            self.refresh(self.offset + 5)

    def _column_selection(self, num, _):
        # If there's a multi-row selection, then mark the
        # whole rows and set self.select_rows so we know
        # to copy the whole rows if <<Copy>> occurs.
        try:
            i1, i2 = self.columns[num].tag_ranges("sel")
            r1 = int(float(str(i1)))
            r2 = int(float(str(i2)))
        except ValueError:
            r1 = 0
            r2 = 0

        for cw in self.columns:
            cw.tag_delete("xsel")

        if r1 != r2:
            self.select_rows = (r1, r2)
            for cw in self.columns:
                cw.tag_add("xsel", f"{r1}.0", f"{r2+1}.0")
                cw.tag_config("xsel", background="lightgrey")
        else:
            self.select_rows = None

    def _column_copy(self, _):
        # If this was a multi-row selection, then replace
        # the copy buffer with a copy of those whole rows.
        if not self.select_rows:
            return  # not multi-row, keep it.

        r1, r2 = self.select_rows
        df = self.dataframe.iloc[self.offset + r1 - 1 : self.offset + r2]
        buf = io.StringIO()
        df.to_csv(buf, sep="\t")

        # XXX very cheesy, but self.clipboard_append() etc didn't
        # seem to work, so this is a terrible workaround ... dump the
        # TSV into a new tk.Text, select the whole thing and copy it
        # into the clipboard.
        top = tk.Toplevel()
        text = tk.Text(top)
        text.insert(tk.END, buf.getvalue())
        text.tag_add("sel", "1.0", tk.END)
        text.event_generate("<<Copy>>")
        top.destroy()



import multiprocessing
import re
import sys
import threading
import tkinter as tk
from tkinter import filedialog, messagebox, ttk
from typing import Optional

import psutil

from countess import VERSION
from countess.core.config import export_config_graphviz, read_config, write_config
from countess.core.pipeline import PipelineGraph
from countess.core.plugins import get_plugin_classes
from countess.gui.config import PluginConfigurator
from countess.gui.logger import LoggerFrame
from countess.gui.mini_browser import MiniBrowserFrame
from countess.gui.tabular import TabularDataFrame
from countess.gui.tree import FlippyCanvas, GraphWrapper
from countess.gui.widgets import info_button
from countess.utils.pandas import concat_dataframes

# import faulthandler
# faulthandler.enable(all_threads=True)


plugin_classes = sorted(get_plugin_classes(), key=lambda x: x.name)


class PluginChooserFrame(tk.Frame):
    def __init__(self, master, title, callback, has_parents, has_children, *a, **k):
        super().__init__(master, *a, **k)

        self.columnconfigure(0, weight=1)

        label_frame = tk.LabelFrame(self, text=title, padx=10, pady=10)
        label_frame.grid(row=1, column=0, sticky=tk.EW, padx=10, pady=10)

        for n, plugin_class in enumerate(plugin_classes):
            if (
                (has_parents and plugin_class.num_inputs == 0)
                or (not has_parents and plugin_class.num_inputs > 0)
                or (has_children and plugin_class.num_outputs == 0)
            ):
                continue

            label_text = plugin_class.description
            tk.Button(
                label_frame,
                text=plugin_class.name,
                command=lambda plugin_class=plugin_class: callback(plugin_class),
            ).grid(row=n + 1, column=0, sticky=tk.EW)
            tk.Label(label_frame, text=label_text).grid(row=n + 1, column=1, sticky=tk.W, padx=10)


class ConfiguratorWrapper:
    """Wraps up the PluginConfigurator or PluginChooserFrame, plus a
    DataFramePreview, to make up the larger part of the main window"""

    config_canvas = None
    config_subframe = None
    config_subframe_id = None
    preview_subframe = None
    config_change_task = None
    notes_widget = None
    node_update_thread = None
    info_toplevel = None
    info_frame = None

    def __init__(self, frame, node, change_callback):
        self.frame = frame
        self.node = node
        self.change_callback = change_callback

        self.name_var = tk.StringVar(self.frame, value=node.name)
        tk.Entry(self.frame, textvariable=self.name_var, font=("TkHeadingFont", 14, "bold")).grid(
            row=0, columnspan=2, sticky=tk.EW, padx=10, pady=5
        )
        self.name_var.trace("w", self.name_changed_callback)

        self.label = tk.Label(self.frame, justify=tk.LEFT, wraplength=500)
        self.label.grid(sticky=tk.EW, row=1, padx=10, pady=5)
        self.label.bind("<Configure>", self.on_label_configure)

        self.show_config_subframe()

        if self.node.plugin:
            if self.node.is_dirty:
                self.config_change_callback()
            else:
                self.show_preview_subframe()

    def show_config_subframe(self):
        if self.config_canvas:
            self.config_canvas.destroy()
        self.config_canvas = tk.Canvas(self.frame)
        self.config_scrollbar = ttk.Scrollbar(self.frame, orient=tk.VERTICAL, command=self.config_canvas.yview)
        self.config_canvas.configure(yscrollcommand=self.config_scrollbar.set, bd=0)
        self.config_canvas.grid(row=3, column=0, sticky=tk.NSEW)
        self.config_scrollbar.grid(row=3, column=1, sticky=tk.NS)

        self.logger_subframe = LoggerFrame(self.frame)
        self.logger = self.logger_subframe.get_logger(self.node.name)

        if self.node.plugin:
            if self.node.notes:
                self.show_notes_widget(self.node.notes)
            else:
                self.notes_widget = tk.Button(self.frame, text="add notes", command=self.on_add_notes)
                self.notes_widget.grid(row=2, columnspan=2, padx=10, pady=5)

            descr = re.sub(r"\s+", " ", self.node.plugin.description)
            if self.node.plugin.additional:
                descr += "\n" + re.sub(r"\s+", " ", self.node.plugin.additional)

            self.label["text"] = "%s %s — %s" % (
                self.node.plugin.name,
                self.node.plugin.version,
                descr,
            )
            if self.node.plugin.link:
                info_button(self.frame, command=self.on_info_button_press).place(anchor=tk.NE, relx=1, y=50)
            # self.node.prepare(self.logger)
            # self.node.plugin.update()
            self.configurator = PluginConfigurator(self.config_canvas, self.node.plugin, self.config_change_callback)
            self.config_subframe = self.configurator.frame
            self.frame.rowconfigure(3, weight=1)
            self.frame.rowconfigure(4, weight=1)

        else:
            has_parents = len(self.node.parent_nodes) > 0
            has_children = len(self.node.child_nodes) > 0
            self.config_subframe = PluginChooserFrame(
                self.config_canvas, "Choose Plugin", self.choose_plugin, has_parents, has_children
            )
            self.config_subframe.grid(sticky=tk.NSEW)
            self.frame.rowconfigure(3, weight=1)
            self.frame.rowconfigure(4, weight=0)

        self.config_subframe_id = self.config_canvas.create_window((0, 0), window=self.config_subframe, anchor=tk.NW)
        self.config_subframe.bind(
            "<Configure>",
            lambda e: self.config_canvas.configure(scrollregion=self.config_canvas.bbox("all")),
        )
        self.config_canvas.bind("<Configure>", self.on_config_canvas_configure)

    def on_config_canvas_configure(self, *_):
        self.config_canvas.itemconfigure(self.config_subframe_id, width=self.config_canvas.winfo_width())

    def on_label_configure(self, *_):
        self.label["wraplength"] = self.label.winfo_width() - 20

    def on_info_button_press(self, *_):
        if self.info_toplevel is None:
            self.info_toplevel = tk.Toplevel()
            self.info_toplevel.protocol("WM_DELETE_WINDOW", self.on_info_toplevel_close)
            self.info_frame = MiniBrowserFrame(self.info_toplevel, self.node.plugin.link)
            self.info_frame.pack(fill="both", expand=True)
        else:
            self.info_frame.load_url(self.node.plugin.link)

    def on_info_toplevel_close(self):
        self.info_toplevel.destroy()
        self.info_toplevel = None

    def on_add_notes(self, *_):
        self.notes_widget.destroy()
        self.show_notes_widget()

    def show_notes_widget(self, notes=""):
        self.notes_widget = tk.Text(self.frame, height=5)
        self.notes_widget.insert("1.0", notes)
        self.notes_widget.bind("<<Modified>>", self.notes_modified_callback)
        self.notes_widget.grid(row=2, columnspan=2, sticky=tk.EW, padx=10, pady=5)

    def show_preview_subframe(self):
        if self.preview_subframe:
            self.preview_subframe.destroy()

        if not self.node.plugin.show_preview:
            self.frame.rowconfigure(4, weight=0)
            return
        elif not self.node.result:
            self.preview_subframe = tk.Frame(self.frame)
            self.preview_subframe.columnconfigure(0, weight=1)
            tk.Label(self.preview_subframe, text="no result").grid(sticky=tk.EW)
        elif all(isinstance(r, (str, bytes)) for r in self.node.result):
            text_result = "".join(self.node.result)
            self.preview_subframe = tk.Frame(self.frame, highlightbackground="black", highlightthickness=3)
            self.preview_subframe.columnconfigure(0, weight=1)
            self.preview_subframe.rowconfigure(1, weight=1)
            n_lines = len(text_result.splitlines())
            tk.Label(self.preview_subframe, text=f"Text Preview {n_lines} Lines").grid(sticky=tk.NSEW)
            text = tk.Text(self.preview_subframe)
            text.insert("1.0", text_result)
            text["state"] = "disabled"
            text.grid(sticky=tk.NSEW)
        else:
            try:
                df = concat_dataframes(self.node.result)
                self.preview_subframe = TabularDataFrame(self.frame, highlightbackground="black", highlightthickness=3)
                self.preview_subframe.set_dataframe(df)
                self.preview_subframe.set_sort_order(self.node.sort_column or 0, self.node.sort_descending)
                self.preview_subframe.set_callback(self.preview_changed_callback)
            except (TypeError, ValueError):
                self.preview_subframe = tk.Frame(self.frame)
                self.preview_subframe.columnconfigure(0, weight=1)
                tk.Label(self.preview_subframe, text="no result").grid(sticky=tk.EW)

        self.preview_subframe.grid(row=4, columnspan=2, sticky=tk.NSEW)

    def preview_changed_callback(self, offset: int, sort_col: int, sort_desc: bool) -> None:
        self.node.sort_column = sort_col
        self.node.sort_descending = sort_desc

    def name_changed_callback(self, *_):
        name = self.name_var.get()
        self.node.name = name
        self.change_callback(self.node)

    def notes_modified_callback(self, *_):
        self.node.notes = self.notes_widget.get("1.0", tk.END).strip()
        self.notes_widget.edit_modified(False)

    def config_change_callback(self, *_):
        self.node.mark_dirty()
        if self.config_change_task:
            self.frame.after_cancel(self.config_change_task)
        self.config_change_task = self.frame.after(1000, self.config_change_task_callback)

    def config_change_task_callback(self):
        self.config_change_task = None

        self.node_update_thread = threading.Thread(target=self.node.prerun, args=(self.logger,))
        self.node_update_thread.start()

        self.logger_subframe.clear()
        self.logger_subframe.grid(row=5, columnspan=2, sticky=tk.NSEW)
        self.logger_subframe.after(100, self.config_change_task_callback_2)
        self.change_callback(self.node)

    def config_change_task_callback_2(self):
        self.logger.poll()

        if self.node_update_thread.is_alive():
            self.logger_subframe.after(100, self.config_change_task_callback_2)
            return

        self.node_update_thread.join()

        # XXX stop the form scrolling away when it is refreshed, by putting
        # it back where it belongs.
        pos1, pos2 = self.config_scrollbar.get()
        self.show_preview_subframe()
        self.configurator.update()
        self.frame.update()
        self.config_canvas.yview_moveto(pos1)
        self.config_scrollbar.set(pos1, pos2)

        if self.logger_subframe.count == 0:
            self.logger_subframe.grid_forget()

    #        else:
    #            self.logger_subframe.progress_hide()

    def choose_plugin(self, plugin_class):
        self.node.plugin = plugin_class()
        self.node.prerun(self.logger)
        self.node.is_dirty = True
        self.show_config_subframe()
        if self.node.name.startswith("NEW "):
            self.node.name = self.node.plugin.name + self.node.name.removeprefix("NEW")
            self.name_var.set(self.node.name)
        self.change_callback(self.node)

    def destroy(self):
        if self.config_change_task:
            self.frame.after_cancel(self.config_change_task)
        if self.config_subframe:
            self.config_subframe.destroy()
        if self.preview_subframe:
            self.preview_subframe.destroy()
        if self.logger_subframe:
            self.logger_subframe.destroy()


class ButtonMenu:  # pylint: disable=R0903
    def __init__(self, tk_parent, buttons):
        self.frame = tk.Frame(tk_parent)
        for button_number, (button_label, button_command) in enumerate(buttons):
            tk.Button(self.frame, text=button_label, command=button_command).grid(
                row=0, column=button_number, sticky=tk.EW
            )
        self.frame.grid(sticky=tk.NSEW)


class RunWindow:
    """Opens a separate window to run the pipeline in.  The actual pipeline is then run
    in a separate process as well, so that it can be stopped."""

    def __init__(self, graph):
        self.graph = graph

        self.toplevel = tk.Toplevel()
        self.toplevel.columnconfigure(0, weight=1)
        self.toplevel.rowconfigure(1, weight=1)

        tk.Label(self.toplevel, text="Run").grid(row=0, column=0, stick=tk.EW)

        self.logger_frame = LoggerFrame(self.toplevel)
        self.logger_frame.grid(row=1, column=0, stick=tk.NSEW)
        self.logger = self.logger_frame.get_logger("")
        self.logger.info("Started")

        self.button = tk.Button(self.toplevel, text="Stop", command=self.on_button)
        self.button.grid(row=2, column=0, sticky=tk.EW)

        self.process = multiprocessing.Process(target=self.subproc)
        self.process.start()

        self.poll()

    def subproc(self):
        self.graph.run(self.logger)

    def poll(self):
        if self.process:
            if self.process.is_alive():
                self.logger_frame.after(1000, self.poll)
            else:
                self.logger.info("Finished")
                self.process = None
                self.button["text"] = "Close"
            self.logger_frame.poll()

    def on_button(self):
        if self.process:
            for p in psutil.Process(self.process.pid).children(recursive=True):
                p.terminate()
            self.process.terminate()
            self.button["text"] = "Close"
            self.process = None

            self.logger.info("Stopped")
            self.logger_frame.poll()

        elif self.toplevel:
            self.toplevel.destroy()
            self.toplevel = None


class MainWindow:
    """Arrange the parts of the main window, with the FrameWrapper on the top
    left and the ConfiguratorWrapper either beside it or under it depending on
    the window dimenstions"""

    graph_wrapper = None
    preview_frame = None
    config_wrapper = None
    config_changed = False

    def __init__(self, tk_parent: tk.Widget, config_filename: Optional[str] = None):
        self.tk_parent = tk_parent

        ButtonMenu(
            tk_parent,
            [
                ("New Config", self.config_new),
                ("Load Config", self.config_load),
                ("Save Config", self.config_save),
                ("Export Config", self.config_export),
                ("Tidy Graph", self.graph_tidy),
                ("Run", self.program_run),
                ("Exit", self.program_exit),
            ],
        )

        self.frame = tk.Frame(tk_parent)
        self.frame.grid(sticky=tk.NSEW)

        # The left (or top) pane, which contains the pipeline graph
        self.canvas = FlippyCanvas(self.frame, bg="skyblue")

        # The right (or bottom) pane, which contains everything else.
        # 0: The node label
        # 1: The plugin description
        # 2: Node notes / add notes button
        # 3: Configuration
        # 4: Preview pane
        # 5: Log output

        self.subframe = tk.Frame(self.frame)
        self.subframe.columnconfigure(0, weight=1)
        self.subframe.columnconfigure(1, weight=0)
        self.subframe.rowconfigure(0, weight=0)
        self.subframe.rowconfigure(1, weight=0)
        self.subframe.rowconfigure(2, weight=0)
        self.subframe.rowconfigure(3, weight=1)
        self.subframe.rowconfigure(4, weight=1)
        self.subframe.rowconfigure(5, weight=0)

        self.frame.bind("<Configure>", self.on_frame_configure, add=True)

        self.logger_subframe = LoggerFrame(self.subframe)
        self.logger_subframe.grid(row=5, columnspan=2, sticky=tk.NSEW)

        if config_filename:
            self.config_load(config_filename)
        else:
            self.config_new()

    def config_new(self):
        if self.config_changed:
            if not messagebox.askokcancel("Clear Config", "Clear unsaved config?"):
                return
        self.config_changed = False
        self.config_filename = None
        if self.graph_wrapper:
            self.graph_wrapper.destroy()
        self.graph = PipelineGraph()
        self.graph_wrapper = GraphWrapper(self.canvas, self.graph, self.node_select)
        self.graph_wrapper.add_new_node(select=True)

    def config_load(self, filename=None):
        if not filename:
            filename = filedialog.askopenfilename(filetypes=[(".INI Config File", "*.ini")])
        if not filename:
            return
        self.config_filename = filename
        if self.graph_wrapper:
            self.graph_wrapper.destroy()
        self.graph = read_config(filename)
        self.graph_wrapper = GraphWrapper(self.canvas, self.graph, self.node_select)
        self.node_select(None)

    def config_save(self, filename=None):
        if not filename:
            filename = filedialog.asksaveasfilename(
                initialfile=self.config_filename,
                filetypes=[(".INI Config File", "*.ini")],
            )
        if not filename:
            return
        write_config(self.graph, filename)
        self.config_changed = False
        # XXX there should be a self.graph_wrapper.refresh()
        # Names may have changed on save
        self.graph_wrapper.destroy()
        self.graph_wrapper = GraphWrapper(self.canvas, self.graph, self.node_select)

    def config_export(self, filename=None):
        if not filename:
            if self.config_filename:
                initialfile = self.config_filename.removesuffix(".ini") + ".dot"
            else:
                initialfile = None
            filename = filedialog.asksaveasfilename(initialfile=initialfile, filetypes=[("Graphviz File", "*.dot")])
        if not filename:
            return
        export_config_graphviz(self.graph, filename)

    def graph_tidy(self):
        self.graph.tidy()
        # XXX there should be a self.graph_wrapper.refresh()
        self.graph_wrapper.destroy()
        self.graph_wrapper = GraphWrapper(self.canvas, self.graph, self.node_select)

    def program_run(self):
        RunWindow(self.graph)

    def program_exit(self):
        if not self.config_changed or messagebox.askokcancel("Exit", "Exit CountESS without saving config?"):
            self.tk_parent.quit()

    def node_select(self, node):
        for widget in self.subframe.winfo_children():
            widget.destroy()
        if node:
            if self.config_wrapper:
                self.config_wrapper.destroy()
            self.config_wrapper = ConfiguratorWrapper(self.subframe, node, self.node_changed)

    def node_changed(self, node):
        self.config_changed = True
        self.graph_wrapper.node_changed(node)

    def on_frame_configure(self, event):
        """Swaps layout around when the window goes from landscape to portrait"""
        if event.width > event.height:
            x = event.width // 4
            self.canvas.place(x=0, y=0, w=x, h=event.height)
            self.subframe.place(x=x, y=0, w=event.width - x, h=event.height)
        else:
            y = event.height // 4
            self.canvas.place(x=0, y=0, w=event.width, h=y)
            self.subframe.place(x=0, y=y, w=event.width, h=event.height - y)


def make_root():
    try:
        import ttkthemes  # pylint: disable=C0415

        root = ttkthemes.ThemedTk()
        themes = set(root.get_themes())
        for t in ["clam", "aqua", "winnative"]:
            if t in themes:
                root.set_theme(t)
                break
    except ImportError:
        root = tk.Tk()
        # XXX some kind of ttk style setup goes here as a fallback

    root.title(f"CountESS {VERSION}")
    root.rowconfigure(0, weight=0)
    root.rowconfigure(1, weight=1)
    root.columnconfigure(0, weight=1)

    # Try to start off maximized, but this option doesn't exist
    # in all Tks or all platforms.
    try:
        root.state("zoomed")
    except tk.TclError:
        pass

    return root


def main():
    root = make_root()
    MainWindow(root, sys.argv[1] if len(sys.argv) > 1 else None)

    root.mainloop()


if __name__ == "__main__":
    main()



from typing import Dict, Iterable, Optional

import pandas as pd

from countess import VERSION
from countess.core.logger import Logger
from countess.core.parameters import ArrayParam, BooleanParam, ColumnOrIndexChoiceParam, MultiParam
from countess.core.plugins import PandasProductPlugin
from countess.utils.pandas import get_all_columns


def _join_how(left_required: bool, right_required: bool) -> str:
    if left_required:
        return "inner" if right_required else "left"
    else:
        return "right" if right_required else "outer"


class JoinPlugin(PandasProductPlugin):
    """Joins Pandas Dataframes"""

    name = "Join"
    description = "Joins two Pandas Dataframes by indexes or columns"
    version = VERSION
    link = "https://countess-project.github.io/CountESS/included-plugins/#join"

    parameters = {
        "inputs": ArrayParam(
            "Inputs",
            MultiParam(
                "Input",
                {
                    "join_on": ColumnOrIndexChoiceParam("Join On"),
                    "required": BooleanParam("Required", True),
                    "drop": BooleanParam("Drop Column", False),
                },
            ),
            read_only=True,
            min_size=2,
            max_size=2,
        ),
    }
    join_params = None
    input_columns_1: Optional[Dict] = None
    input_columns_2: Optional[Dict] = None

    def prepare(self, sources: list[str], row_limit: Optional[int] = None):
        super().prepare(sources, row_limit)

        assert isinstance(self.parameters["inputs"], ArrayParam)
        assert len(self.parameters["inputs"]) == 2
        ip1, ip2 = self.parameters["inputs"]
        ip1.label = f"Input 1: {sources[0]}"
        ip2.label = f"Input 2: {sources[1]}"

        self.join_params = {
            "how": _join_how(ip1.required.value, ip2.required.value),
            "left_index": ip1.join_on.is_index(),
            "right_index": ip2.join_on.is_index(),
            "left_on": None if ip1.join_on.is_index() else ip1.join_on.value,
            "right_on": None if ip2.join_on.is_index() else ip2.join_on.value,
        }
        self.input_columns_1 = {}
        self.input_columns_2 = {}

    def process_dataframes(self, dataframe1: pd.DataFrame, dataframe2: pd.DataFrame, logger: Logger) -> pd.DataFrame:
        # update columns on inputs, these won't propagate back in the case of multiprocess runs but
        # they will work in preview mode where we only run this in a single thread.

        assert self.input_columns_1 is not None
        assert self.input_columns_2 is not None
        assert self.join_params is not None
        assert isinstance(self.parameters["inputs"], ArrayParam)

        self.input_columns_1.update(get_all_columns(dataframe1))
        self.input_columns_2.update(get_all_columns(dataframe2))

        # "left_on" and "right_on" don't seem to mind if the column
        # is an index, but don't seem to work correctly if the column
        # is part of a multiindex: the other multiindex columns go missing.
        join1 = self.join_params.get("left_on")
        if join1 and dataframe1.index.name != join1:
            drop_index = dataframe1.index.name is None and dataframe1.index.names[0] is None
            dataframe1 = dataframe1.reset_index(drop=drop_index)

        join2 = self.join_params.get("right_on")
        if join2 and dataframe2.index.name != join2:
            drop_index = dataframe2.index.name is None and dataframe2.index.names[0] is None
            dataframe2 = dataframe2.reset_index(drop=drop_index)

        try:
            dataframe = dataframe1.merge(dataframe2, **self.join_params)
        except ValueError as exc:
            logger.exception(exc)
            return pd.DataFrame()

        if self.parameters["inputs"][0]["drop"].value and join1 in dataframe.columns:
            dataframe.drop(columns=join1, inplace=True)
        if self.parameters["inputs"][1]["drop"].value and join2 in dataframe.columns:
            dataframe.drop(columns=join2, inplace=True)

        return dataframe

    def finalize(self, logger: Logger) -> Iterable:
        assert isinstance(self.parameters["inputs"], ArrayParam)
        assert len(self.parameters["inputs"]) == 2
        assert self.input_columns_1 is not None
        assert self.input_columns_2 is not None
        ip1, ip2 = self.parameters["inputs"]

        ip1.set_column_choices(self.input_columns_1.keys())
        ip2.set_column_choices(self.input_columns_2.keys())
        yield from super().finalize(logger)



from itertools import islice, product
from typing import Iterable, Optional

import pandas as pd

from countess import VERSION
from countess.core.logger import Logger
from countess.core.parameters import BooleanParam, StringCharacterSetParam
from countess.core.plugins import PandasInputPlugin


def mutagenize(
    sequence: str, mutate: bool, delete: bool, del3: bool, insert: bool, ins3: bool
) -> Iterable[tuple[str, int, Optional[str], Optional[str]]]:
    # XXX not really happy with how the args are multiplying here!
    # XXX it'd be faster, but less neat, to include logic for duplicate
    # removal here instead of producing duplicates and then removing them
    # later.
    for n, b1 in enumerate(sequence):
        for b2 in "ACGT":
            if mutate and b1 != b2:
                yield sequence[0:n] + b2 + sequence[n + 1 :], n + 1, b1, b2
            if insert:
                yield sequence[0:n] + b2 + sequence[n:], n + 1, None, b2
        if delete:
            yield sequence[0:n] + sequence[n + 1 :], n + 1, b1, None
        if del3 and n + 3 <= len(sequence):
            yield sequence[0:n] + sequence[n + 3 :], n + 1, sequence[n : n + 3], None
        if ins3:
            for ins in product("ACGT", "ACGT", "ACGT"):
                ins_str = "".join(ins)
                yield sequence[0:n] + ins_str + sequence[n:], n + 1, None, ins_str

    ll = len(sequence) + 1
    if insert:
        for b2 in "ACGT":
            yield sequence + b2, ll, None, b2
    if ins3:
        for ins in product("ACGT", "ACGT", "ACGT"):
            ins_str = "".join(ins)
            yield sequence + ins_str, ll, None, ins_str


class MutagenizePlugin(PandasInputPlugin):
    """Mutagenize"""

    name = "Mutagenize"
    description = "Provides all mutations of a sequence"
    link = "https://countess-project.github.io/CountESS/included-plugins/#mutagenize"
    version = VERSION

    character_set = set(("A", "C", "G", "T", "N"))

    parameters = {
        "sequence": StringCharacterSetParam("Sequence", "", character_set=character_set),
        "mutate": BooleanParam("All Single Mutations?", True),
        "delete": BooleanParam("All Single Deletes?", False),
        "del3": BooleanParam("All Triple Deletes?", False),
        "insert": BooleanParam("All Single Inserts?", False),
        "ins3": BooleanParam("All Triple Inserts?", False),
        "remove": BooleanParam("Remove Duplicates?", False),
    }

    def num_files(self):
        return 1

    def load_file(self, file_number: int, logger: Logger, row_limit: Optional[int] = None) -> Iterable[pd.DataFrame]:
        assert file_number == 0

        df = pd.DataFrame(
            islice(
                mutagenize(
                    self.parameters["sequence"].value,
                    self.parameters["mutate"].value,
                    self.parameters["delete"].value,
                    self.parameters["del3"].value,
                    self.parameters["insert"].value,
                    self.parameters["ins3"].value,
                ),
                0,
                row_limit,
            ),
            columns=["sequence", "position", "reference", "variation"],
        )
        if self.parameters["remove"].value:
            df = df.groupby(["sequence"]).agg({"sequence": "count"}).rename(columns={"sequence": "count"})
        yield df



import time
from queue import Empty, Queue
from threading import Thread
from typing import Any, Optional

from countess.core.logger import Logger
from countess.core.plugins import BasePlugin, FileInputPlugin, ProcessPlugin, get_plugin_classes

PRERUN_ROW_LIMIT = 100000


class SentinelQueue(Queue):

    """This is an easy and a bit lazy way of making a queue iterable.
    The writer is expected to call `queue.finish()` when it is done and
    the reader can treat the queue like an iterable."""

    # XXX this is an attempt to handle multiple threads reading from the
    # queue in parallel: they should all get StopIterations.
    stopped = False

    class SENTINEL:
        pass

    def finish(self):
        self.put(self.SENTINEL)

    def __iter__(self):
        return self

    def get_or_raise_on_stopped(self):
        while True:
            if self.stopped:
                raise StopIteration
            try:
                return super().get(timeout=1)
            except Empty:
                pass

    def __next__(self):
        val = self.get_or_raise_on_stopped()
        if val is self.SENTINEL:
            self.stopped = True
            raise StopIteration
        return val

    def get(self, block=True, timeout=None):
        if self.stopped:
            raise ValueError("SentinelQueue stopped")
        return super().get(block, timeout)

    def put(self, item, block=True, timeout=None):
        if self.stopped:
            raise ValueError("SentinelQueue stopped")
        super().put(item, block, timeout)


class PipelineNode:
    name: str
    plugin: Optional[BasePlugin] = None
    position: Optional[tuple[float, float]] = None
    sort_column: int = 0
    sort_descending: bool = False
    notes: Optional[str] = None
    parent_nodes: set["PipelineNode"]
    child_nodes: set["PipelineNode"]
    config: Optional[list[tuple[str, str, str]]] = None
    result: Any = None
    is_dirty: bool = True

    output_queues: set[SentinelQueue]
    counter_in: int = 0
    counter_out: int = 0

    # XXX config is a cache for config loaded from the file
    # at config load time, if it is present it is loaded the
    # first time the plugin is prerun.

    def __init__(self, name, plugin=None, config=None, position=None, notes=None, sort_column=0, sort_descending=0):
        self.name = name
        self.plugin = plugin
        self.config = config or []
        self.position = position
        self.sort_column = sort_column
        self.sort_descending = sort_descending
        self.notes = notes
        self.parent_nodes = set()
        self.child_nodes = set()
        self.output_queues = set()

    def __hash__(self):
        return id(self)

    def is_ancestor_of(self, node):
        return (self in node.parent_nodes) or any((self.is_ancestor_of(n) for n in node.parent_nodes))

    def is_descendant_of(self, node):
        return (self in node.child_nodes) or any((self.is_descendant_of(n) for n in node.child_nodes))

    def plugin_process(self, x):
        self.plugin.process(*x)

    def add_output_queue(self):
        queue = SentinelQueue(maxsize=3)
        self.output_queues.add(queue)
        return queue

    def clear_output_queues(self):
        self.output_queues = set()

    def queue_output(self, result):
        for data in result:
            self.counter_out += 1
            # XXX can we do this out-of-order if any queues are full?
            for queue in self.output_queues:
                queue.put(data)

    def finish_output(self):
        for queue in self.output_queues:
            queue.finish()

    def run_multithread(self, queue: SentinelQueue, name: str, logger: Logger, row_limit: Optional[int] = None):
        assert isinstance(self.plugin, ProcessPlugin)
        for data_in in queue:
            self.counter_in += 1
            self.queue_output(self.plugin.process(data_in, name, logger))

    def run_subthread(self, queue: SentinelQueue, name: str, logger: Logger, row_limit: Optional[int] = None):
        assert isinstance(self.plugin, ProcessPlugin)

        for data_in in queue:
            self.counter_in += 1
            self.queue_output(self.plugin.process(data_in, name, logger))
        self.queue_output(self.plugin.finished(name, logger))

    def run_thread(self, logger: Logger, row_limit: Optional[int] = None):
        """For each PipelineNode, this is run in its own thread."""
        assert isinstance(self.plugin, (ProcessPlugin, FileInputPlugin))

        self.plugin.prepare([node.name for node in self.parent_nodes], row_limit)

        if len(self.parent_nodes) == 1:
            assert isinstance(self.plugin, ProcessPlugin)
            # there is only a single parent node, run several subthreads to
            # do the processing
            only_parent_node = list(self.parent_nodes)[0]
            only_parent_queue = only_parent_node.add_output_queue()
            subthreads = [
                Thread(target=self.run_multithread, args=(only_parent_queue, only_parent_node.name, logger, row_limit))
                for _ in range(0, 4)
            ]
            for subthread in subthreads:
                subthread.start()
            for subthread in subthreads:
                subthread.join()

            self.queue_output(self.plugin.finished(only_parent_node.name, logger))

        elif len(self.parent_nodes) > 1:
            assert isinstance(self.plugin, ProcessPlugin)
            # there are multiple parent nodes: spawn off a subthread to handle
            # each of them.
            subthreads = [
                Thread(
                    target=self.run_subthread,
                    args=(parent_node.add_output_queue(), parent_node.name, logger, row_limit),
                )
                for parent_node in self.parent_nodes
            ]
            for subthread in subthreads:
                subthread.start()
            for subthread in subthreads:
                subthread.join()

        self.queue_output(self.plugin.finalize(logger))
        self.finish_output()

    def load_config(self, logger: Logger):
        assert isinstance(self.plugin, BasePlugin)
        if self.config:
            for key, val, base_dir in self.config:
                try:
                    self.plugin.set_parameter(key, val, base_dir)
                except (KeyError, ValueError):
                    logger.warning(f"Parameter {key}={val} Not Found")
            self.config = None

    def prerun(self, logger: Logger, row_limit=PRERUN_ROW_LIMIT):
        if not self.plugin:
            return
        self.load_config(logger)
        if self.is_dirty:
            assert isinstance(self.plugin, (ProcessPlugin, FileInputPlugin))
            self.result = []
            self.plugin.prepare([node.name for node in self.parent_nodes], row_limit)

            for parent_node in self.parent_nodes:
                assert isinstance(self.plugin, ProcessPlugin)
                parent_node.prerun(logger, row_limit)
                if parent_node.result:
                    for data_in in parent_node.result:
                        self.result += list(self.plugin.process(data_in, parent_node.name, logger))
                self.result += list(self.plugin.finished(parent_node.name, logger))
            self.result += list(self.plugin.finalize(logger))
            self.is_dirty = False

    def mark_dirty(self):
        self.is_dirty = True
        for child_node in self.child_nodes:
            if not child_node.is_dirty:
                child_node.mark_dirty()

    def add_parent(self, parent):
        if (not self.plugin or self.plugin.num_inputs) and (not parent.plugin or parent.plugin.num_outputs):
            self.parent_nodes.add(parent)
            parent.child_nodes.add(self)
            self.mark_dirty()

    def del_parent(self, parent):
        self.parent_nodes.discard(parent)
        parent.child_nodes.discard(self)
        self.mark_dirty()

    def has_sibling(self):
        return any(len(pn.child_nodes) > 1 for pn in self.parent_nodes)

    def configure_plugin(self, key, value, base_dir="."):
        self.plugin.set_parameter(key, value, base_dir)
        self.mark_dirty()

    def final_descendants(self):
        if self.child_nodes:
            return set(n2 for n1 in self.child_nodes for n2 in n1.final_descendants())
        else:
            return set(self)

    def detatch(self):
        for parent_node in self.parent_nodes:
            parent_node.child_nodes.discard(self)
        for child_node in self.child_nodes:
            child_node.parent_nodes.discard(self)

    @classmethod
    def get_ancestor_list(cls, nodes):
        """Given a bunch of nodes, find the list of all the ancestors in a
        sensible order"""
        parents = set((p for n in nodes for p in n.parent_nodes))
        if not parents:
            return list(nodes)
        return cls.get_ancestor_list(parents) + list(nodes)


class PipelineGraph:
    def __init__(self):
        self.plugin_classes = get_plugin_classes()
        self.nodes = []

    def add_node(self, node):
        self.nodes.append(node)

    def del_node(self, node):
        node.detatch()
        self.nodes.remove(node)

    def traverse_nodes(self):
        found_nodes = set(node for node in self.nodes if not node.parent_nodes)
        yield from found_nodes

        while len(found_nodes) < len(self.nodes):
            for node in self.nodes:
                if node not in found_nodes and node.parent_nodes.issubset(found_nodes):
                    yield node
                    found_nodes.add(node)

    def traverse_nodes_backwards(self):
        found_nodes = set(node for node in self.nodes if not node.child_nodes)
        yield from found_nodes

        while len(found_nodes) < len(self.nodes):
            for node in self.nodes:
                if node not in found_nodes and node.child_nodes.issubset(found_nodes):
                    yield node
                    found_nodes.add(node)

    def run(self, logger):
        threads_and_nodes = []
        for node in self.traverse_nodes_backwards():
            node.load_config(logger)
            threads_and_nodes.append((Thread(target=node.run_thread, args=(logger,)), node))

        for thread, _ in threads_and_nodes:
            thread.start()

        while True:
            print("------------------")
            for thread, node in threads_and_nodes[::-1]:
                if thread.is_alive():
                    print("%-40s %d %d" % (node.name, node.counter_in, node.counter_out))
            if not any(t.is_alive() for t, _ in threads_and_nodes):
                break
            time.sleep(10)

    def reset(self):
        for node in self.nodes:
            node.result = None
            node.is_dirty = True

    def tidy(self):
        """Tidies the graph (sets all the node positions)"""

        # XXX This is very arbitrary and not particularly efficient.
        # Some kind of FDP-like algorithm might be nice.
        # Especially if it could include node/line collisions.
        # See #24

        nodes = list(self.traverse_nodes())

        # first calculate a stratum for each node.

        stratum = {}
        for node in nodes:
            if not node.parent_nodes:
                stratum[node] = 0
            else:
                stratum[node] = max(stratum[n] for n in node.parent_nodes) + 1

        # shufffle nodes back down to avoid really long connections.

        for node in nodes[::-1]:
            if node.child_nodes:
                if len(node.parent_nodes) == 0:
                    stratum[node] = min(stratum[n] for n in node.child_nodes) - 1
                else:
                    stratum[node] = (
                        min(stratum[n] for n in node.child_nodes) + max(stratum[n] for n in node.parent_nodes)
                    ) // 2

        max_stratum = max(stratum.values())

        position = {}
        for s in range(0, max_stratum + 1):
            # now sort all the nodes by the average position of their parents,
            # to try and stop them forming a big tangle.  The current position
            # is included as a "tie breaker" and to keep some memory of the user's
            # preference for position (eg: ordering of branches)

            def avg_pos_parents(node):
                return sum(position[p] for p in node.parent_nodes) / len(node.parent_nodes)

            snodes = [
                (
                    avg_pos_parents(node) if node.parent_nodes else 0.5,
                    node.position[1],
                    n,
                )
                for n, node in enumerate(nodes)
                if stratum[node] == s
            ]
            snodes.sort()

            # Assign node positions with the stratums placed
            # evenly and the nodes spaced evenly per stratum.

            y = (s + 0.5) / (max_stratum + 1)
            for p, (_, _, n) in enumerate(snodes):
                x = (p + 0.5) / len(snodes)
                nodes[n].position = (y, x)
                position[nodes[n]] = x



import csv
import gzip
from io import BufferedWriter, BytesIO
from typing import Optional, Union

import pandas as pd

from countess import VERSION
from countess.core.logger import Logger
from countess.core.parameters import (
    ArrayParam,
    BooleanParam,
    ChoiceParam,
    DataTypeOrNoneChoiceParam,
    FileSaveParam,
    MultiParam,
    StringParam,
)
from countess.core.plugins import PandasInputFilesPlugin, PandasOutputPlugin
from countess.utils.pandas import flatten_columns


class LoadCsvPlugin(PandasInputFilesPlugin):
    """Load CSV files"""

    name = "CSV Load"
    description = "Loads data from CSV or similar delimited text files and assigns types to columns"
    link = "https://countess-project.github.io/CountESS/included-plugins/#csv-reader"
    version = VERSION

    file_types = [("CSV", [".csv", ".gz"]), ("TSV", [".tsv", ".gz"]), ("TXT", ".txt")]

    parameters = {
        "delimiter": ChoiceParam("Delimiter", ",", choices=[",", ";", "TAB", "|", "WHITESPACE"]),
        "quoting": ChoiceParam("Quoting", "None", choices=["None", "Double-Quote", "Quote with Escape"]),
        "comment": ChoiceParam("Comment", "None", choices=["None", "#", ";"]),
        "header": BooleanParam("CSV file has header row?", True),
        "filename_column": StringParam("Filename Column", ""),
        "columns": ArrayParam(
            "Columns",
            MultiParam(
                "Column",
                {
                    "name": StringParam("Column Name", ""),
                    "type": DataTypeOrNoneChoiceParam("Column Type"),
                    "index": BooleanParam("Index?", False),
                },
            ),
        ),
    }

    def read_file_to_dataframe(self, file_params, logger, row_limit=None):
        filename = file_params["filename"].value

        options = {
            "header": 0 if self.parameters["header"].value else None,
        }
        if row_limit is not None:
            options["nrows"] = row_limit

        index_col_numbers = []

        if len(self.parameters["columns"]):
            options["names"] = []
            options["dtype"] = {}
            options["usecols"] = []

            for n, pp in enumerate(self.parameters["columns"]):
                options["names"].append(pp["name"].value or f"column_{n}")
                if pp["type"].is_not_none():
                    if pp["index"].value:
                        index_col_numbers.append(len(options["usecols"]))
                    options["usecols"].append(n)
                    options["dtype"][n] = pp["type"].get_selected_type()

        delimiter = self.parameters["delimiter"].value
        if delimiter == "TAB":
            options["delimiter"] = "\t"
        elif delimiter == "WHITESPACE":
            options["delim_whitespace"] = True
        else:
            options["delimiter"] = delimiter

        quoting = self.parameters["quoting"].value
        if quoting == "None":
            options["quoting"] = csv.QUOTE_NONE
        elif quoting == "Double-Quote":
            options["quotechar"] = '"'
            options["doublequote"] = True
        elif quoting == "Quote with Escape":
            options["quotechar"] = '"'
            options["doublequote"] = False
            options["escapechar"] = "\\"

        comment = self.parameters["comment"].value
        if comment != "None":
            options["comment"] = comment

        # XXX pd.read_csv(index_col=) is half the speed of pd.read_csv().set_index()

        df = pd.read_csv(filename, **options)

        while len(df.columns) > len(self.parameters["columns"]):
            self.parameters["columns"].add_row()

        if self.parameters["header"].value:
            for n, col in enumerate(df.columns):
                if not self.parameters["columns"][n]["name"].value:
                    self.parameters["columns"][n]["name"].value = str(col)
                    self.parameters["columns"][n]["type"].value = "string"

        filename_column = self.parameters["filename_column"].value
        if filename_column:
            df[filename_column] = filename

        if index_col_numbers:
            df = df.set_index([df.columns[n] for n in index_col_numbers])

        return df


class SaveCsvPlugin(PandasOutputPlugin):
    name = "CSV Save"
    description = "Save data as CSV or similar delimited text files"
    link = "https://countess-project.github.io/CountESS/included-plugins/#csv-writer"
    version = VERSION

    file_types = [("CSV", [".csv", ".gz"]), ("TSV", [".tsv", ".gz"]), ("TXT", ".txt")]

    parameters = {
        "header": BooleanParam("CSV header row?", True),
        "filename": FileSaveParam("Filename", file_types=file_types),
        "delimiter": ChoiceParam("Delimiter", ",", choices=[",", ";", "TAB", "|", "SPACE"]),
        "quoting": BooleanParam("Quote all Strings", False),
    }

    filehandle: Optional[Union[BufferedWriter, BytesIO]] = None
    csv_columns = None

    SEPARATORS = {",": ",", ";": ";", "SPACE": " ", "TAB": "\t"}
    QUOTING = {False: csv.QUOTE_MINIMAL, True: csv.QUOTE_NONNUMERIC}

    def prepare(self, sources: list[str], row_limit: Optional[int] = None):
        if row_limit is None:
            filename = self.parameters["filename"].value
            if filename.endswith(".gz"):
                self.filehandle = gzip.open(filename, "wb")
            else:
                self.filehandle = open(filename, "wb")
        else:
            self.filehandle = BytesIO()

        self.csv_columns = None

    def process(self, data: pd.DataFrame, source: str, logger: Logger):
        # reset indexes so we can treat all columns equally.
        # if there's just a nameless index then we don't care about it, drop it.
        drop_index = data.index.name is None and data.index.names[0] is None
        dataframe = flatten_columns(data.reset_index(drop=drop_index))

        # if this is our first dataframe to write then decide whether to
        # include the header or not.
        if self.csv_columns is None:
            self.csv_columns = list(dataframe.columns)
            emit_header = bool(self.parameters["header"].value)
        else:
            # add in any columns we haven't seen yet in previous dataframes.
            for c in dataframe.columns:
                if c not in self.csv_columns:
                    self.csv_columns.append(c)
                    logger.warning(f"Added CSV Column {repr(c)} with no header")
            # fill in blanks for any columns which are in previous dataframes but not
            # in this one.
            dataframe = dataframe.assign(**{c: None for c in self.csv_columns if c not in dataframe.columns})
            emit_header = False

        dataframe.to_csv(
            self.filehandle,
            header=emit_header,
            columns=self.csv_columns,
            index=False,
            sep=self.SEPARATORS[self.parameters["delimiter"].value],
            quoting=self.QUOTING[self.parameters["quoting"].value],
        )  # type: ignore [call-overload]
        return []

    def finalize(self, logger: Logger):
        if isinstance(self.filehandle, BytesIO):
            yield self.filehandle.getvalue().decode("utf-8")



from typing import Optional

import pandas as pd

from countess import VERSION
from countess.core.logger import Logger
from countess.core.parameters import BooleanParam, ColumnChoiceParam, ColumnOrNoneChoiceParam, IntegerParam, StringParam
from countess.core.plugins import PandasTransformDictToDictPlugin
from countess.utils.variant import find_variant_string


class VariantPlugin(PandasTransformDictToDictPlugin):
    """Turns a DNA sequence into a HGVS variant code"""

    name = "Variant Translator"
    description = "Turns a DNA sequence into a HGVS variant code"
    version = VERSION
    link = "https://countess-project.github.io/CountESS/included-plugins/#variant-caller"

    parameters = {
        "column": ColumnChoiceParam("Input Column", "sequence"),
        "reference": ColumnOrNoneChoiceParam("Reference Column"),
        "sequence": StringParam("*OR* Reference Sequence"),
        "output": StringParam("Output Column", "variant"),
        "max_mutations": IntegerParam("Max Mutations", 10),
        "protein": StringParam("Protein Column", ""),
        "offset": IntegerParam("Protein Offset", 0),
        "max_protein": IntegerParam("Max Protein Variations", 10),
        "drop": BooleanParam("Drop unidentified variants", False),
        "drop_columns": BooleanParam("Drop Input Column(s)", False),
    }

    def process_dict(self, data, logger: Logger) -> dict:
        assert isinstance(self.parameters["reference"], ColumnOrNoneChoiceParam)
        sequence = data[self.parameters["column"].value]
        if not sequence:
            return {}

        if self.parameters["reference"].is_none():
            reference = self.parameters["sequence"].value
        else:
            reference = data[self.parameters["reference"].value]

        r = {}

        if self.parameters["output"].value:
            try:
                max_mutations = self.parameters["max_mutations"].value
                r[self.parameters["output"].value] = find_variant_string("g.", reference, sequence, max_mutations, 0)
            except ValueError:
                pass
            except (TypeError, KeyError, IndexError) as exc:
                logger.exception(exc)

        if self.parameters["protein"].value:
            try:
                offset = self.parameters["offset"].value
                max_protein = self.parameters["max_protein"].value
                r[self.parameters["protein"].value] = find_variant_string(
                    "p.", reference, sequence, max_protein, offset
                )
            except ValueError:
                pass
            except (TypeError, KeyError, IndexError) as exc:
                logger.exception(exc)

        return r

    def process_dataframe(self, dataframe: pd.DataFrame, logger: Logger) -> Optional[pd.DataFrame]:
        assert isinstance(self.parameters["reference"], ColumnOrNoneChoiceParam)
        df_out = super().process_dataframe(dataframe, logger)

        if df_out is not None:
            if self.parameters["drop"].value:
                if self.parameters["output"].value:
                    df_out.dropna(subset=self.parameters["output"].value, inplace=True)
                if self.parameters["protein"].value:
                    df_out.dropna(subset=self.parameters["protein"].value, inplace=True)
            if self.parameters["drop_columns"].value:
                try:
                    df_out.drop(columns=self.parameters["column"].value, inplace=True)
                    if self.parameters["reference"].is_not_none():
                        df_out.drop(columns=self.parameters["reference"].value, inplace=True)
                except KeyError:
                    pass

        return df_out
